from io import StringIO
import re

import pandas as pd

from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import CsvReader, EMPTY_STR_OR_CSV_LINE
from allotropy.parsers.utils.pandas import df_to_series_data, read_csv, SeriesData
from allotropy.parsers.utils.values import assert_not_none


class CtlImmunospotReader:
    SUPPORTED_EXTENSIONS = "txt"

    header: SeriesData
    plate_identifier: str | None
    plate_data: dict[str, pd.DataFrame]
    histograms: dict[str, tuple[list[float], list[float]]]

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        reader = CsvReader.create(named_file_contents)
        self.header = self._read_header(reader)

        software_version = str(
            assert_not_none(
                re.match(
                    r"^ImmunoSpot ([\d\.]+)$", self.header[str, "Software version"]
                ),
                msg="Unable to parse software version",
            ).group(1)
        )

        if software_version.startswith("7.0.30"):
            (
                self.plate_data,
                self.histograms,
                self.plate_identifier,
            ) = self._parse_7_0_30(reader)
        elif software_version.startswith("7.0.38"):
            (
                self.plate_data,
                self.histograms,
                self.plate_identifier,
            ) = self._parse_7_0_38(reader)
        else:
            msg = f"Unsupported software version: {software_version}"
            raise AllotropeConversionError(msg)

    def _parse_7_0_30(
        self, reader: CsvReader
    ) -> tuple[
        dict[str, pd.DataFrame],
        dict[str, tuple[list[float], list[float]]],
        str | None,
    ]:
        reader.drop_until_inclusive("Unprocessed Data$")

        reader.drop_empty()
        plate_code_line = reader.get() or ""
        assert_not_none(
            re.search("Plate Code =", plate_code_line),
            msg="Unable to find ImmunoSpot Plate Code line",
        )

        plate_identifier = (
            match.group(1)
            if (match := re.search(r"Plate Code = ([\w ]+)", plate_code_line))
            else None
        )
        reader.drop_empty()

        plates: dict[str, pd.DataFrame] = {}
        while reader.current_line_exists():
            raw_name = assert_not_none(
                reader.pop(),
                msg="Unable to read name of assay data.",
            )
            name = raw_name.strip()
            if "Plate Code" in name:
                name = "Spot Count"

            # Read in the plate data, detect end by checking for plate letter label.
            columns = assert_not_none(reader.pop(), "Unable to get column header line")
            reader.pop()
            lines = [columns]
            expected_char = "A"
            while line := reader.pop():
                if line.split()[0] != expected_char:
                    break
                lines.append(line)
                expected_char = chr(ord(expected_char) + 1)

            plates[name] = read_csv(
                StringIO("\n".join(lines)),
                sep=r"\s+",
                # Prevent pandas from rounding decimal values, at the cost of some speed.
                float_precision="round_trip",
                index_col=0,
            )
            reader.drop_until_empty()
            reader.drop_empty()

        # TODO: check all plates are the same size
        return plates, {}, plate_identifier

    def _parse_7_0_38(
        self, reader: CsvReader
    ) -> tuple[
        dict[str, pd.DataFrame],
        dict[str, tuple[list[float], list[float]]],
        str | None,
    ]:
        reader.drop_empty()

        plates: dict[str, pd.DataFrame] = {}
        histograms: dict[str, tuple[list[float], list[float]]] = {}
        while reader.current_line_exists():
            raw_name = assert_not_none(
                reader.pop(),
                msg="Unable to read name of assay data.",
            )
            names = [n.strip() for n in raw_name.strip().split("\t\t") if n.strip()]
            if len(names) == 1 and "Histogram" in names[0]:
                # We currently only handle one histogram, which does not have a label. Raise an error is we get
                # a second one, as we will need to investigate the use case.
                if histograms:
                    msg = "Got unexpected second histogram in input file."
                    raise AllotropeConversionError(msg)
                dimensions, histogram_data = self._read_histogram_7_0_38(reader)
                for well_location, data in histogram_data.items():
                    if well_location in histograms:
                        msg = f"Got multiple histogram entries for well location: {well_location}"
                        raise AllotropeConversionError(msg)
                    histograms[well_location] = (dimensions, data)
            else:
                for name, plate in zip(
                    names, self._read_plate_data_7_0_38(reader), strict=True
                ):
                    plates[name] = plate

            reader.drop_until_empty()
            reader.drop_empty()

        # TODO: check all plates are the same size
        return plates, histograms, None

    def _read_histogram_7_0_38(
        self, reader: CsvReader
    ) -> tuple[list[float], dict[str, list[float]]]:
        dimensions_line = assert_not_none(
            reader.pop(), "Unable to read histogram dimensions line"
        )
        dims = [float(dim) for dim in dimensions_line.strip().split("\t")]
        reader.pop()
        reader.pop()
        histograms: dict[str, list[float]] = {}
        while (line := reader.pop()) is not None:
            if line.startswith("\t"):
                break
            values = line.strip().split("\t")
            histograms[values[0]] = [float(dim) for dim in values[1:]]
        return dims, histograms

    def _read_plate_data_7_0_38(self, reader: CsvReader) -> list[pd.DataFrame]:
        # Read in the plate data, detect end by checking for plate letter label.
        reader.pop()
        columns_line = assert_not_none(reader.pop(), "Unable to get column header line")
        columns = [
            line_split.strip()
            for line_split in columns_line.strip().split("\t\t")
            if line_split.strip()
        ]
        reader.pop()
        reader.pop()
        lines: list[list[str]] = [columns]
        expected_char = "A"
        while line := reader.pop():
            if line.split()[0] != expected_char:
                break
            lines.append(
                [
                    line_split.strip()
                    for line_split in line.strip().split("\t\t")
                    if line_split.strip()
                ]
            )
            expected_char = chr(ord(expected_char) + 1)

        plates: list[pd.DataFrame] = []
        for idx in range(len(lines[0])):
            plates.append(
                read_csv(
                    StringIO("\n".join(line_set[idx] for line_set in lines)),
                    sep=r"\s+",
                    # Prevent pandas from rounding decimal values, at the cost of some speed.
                    float_precision="round_trip",
                    index_col=0,
                )
            )

        return plates

    def _read_header(self, reader: CsvReader) -> SeriesData:
        lines = [line.strip() for line in reader.pop_until_empty(EMPTY_STR_OR_CSV_LINE)]

        def fix_line(line: str) -> str:
            # Add missing key for file path line.
            if line.endswith(".txt") or line.endswith(".xls"):
                line = f"File path: {line}"
            if line.startswith("QC review last updated"):
                date_str = line.split("on ")[1].split(" by")[0]
                line = f"Review Date: {date_str}"
            return line.strip()

        lines = [fix_line(line) for raw_line in lines for line in raw_line.split(";")]
        reader.drop_empty(EMPTY_STR_OR_CSV_LINE)
        if ":" in (reader.get() or ""):
            lines.extend(list(reader.pop_until_empty(EMPTY_STR_OR_CSV_LINE)))

        df = read_csv(
            StringIO("\n".join(lines)),
            sep=r"^([^:]+):\s+",
            header=None,
            engine="python",
            index_col=1,
        ).T
        return df_to_series_data(df, index=-1)



from __future__ import annotations

from allotropy.allotrope.models.adm.pcr.benchling._2023._09.dpcr import Model
from allotropy.allotrope.schema_mappers.adm.pcr.BENCHLING._2023._09.dpcr import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.appbio_absolute_q.appbio_absolute_q_reader import (
    AppbioAbsoluteQReader,
)
from allotropy.parsers.appbio_absolute_q.appbio_absolute_q_structure import (
    create_calculated_data,
    create_measurement_groups,
    create_metadata,
    Group,
    Well,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class AppbioAbsoluteQParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "AppBio AbsoluteQ"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = AppbioAbsoluteQReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = AppbioAbsoluteQReader(named_file_contents)
        wells = Well.create_wells(reader.data)
        groups = Group.create_rows(reader.data)

        return Data(
            create_metadata(
                wells[0].items[0].instrument_identifier,
                named_file_contents.original_file_path,
            ),
            create_measurement_groups(wells),
            calculated_data=create_calculated_data(
                wells, groups, reader.common_columns
            ),
        )



from __future__ import annotations

import logging
from pathlib import Path
import re

from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    ImageFeature,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ProcessedData,
)
from allotropy.exceptions import AllotropeParsingError
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.mabtech_apex import constants
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import assert_not_none


def create_metadata(plate_info: SeriesData, file_path: str) -> Metadata:
    machine_id = assert_not_none(
        re.match(
            "([A-Z]+[a-z]*) ([0-9]+)",
            plate_info[str, "Machine ID:"],
        ),
        msg="Unable to interpret Machine ID",
    )

    return Metadata(
        device_identifier=NOT_APPLICABLE,
        software_name=constants.SOFTWARE_NAME,
        software_version=plate_info.get(str, "Software Version:"),
        model_number=machine_id.group(1),
        equipment_serial_number=machine_id.group(2),
        file_name=Path(file_path).name,
        unc_path=plate_info.get(str, "Path:", file_path),
        custom_info=plate_info.get_unread(),
    )


def _create_measurement(plate_data: SeriesData) -> Measurement:
    location_id = plate_data[str, "Well"]
    well_plate = plate_data.get(str, "Plate")
    led_filter = plate_data.get(str, "LED Filter")
    led_filter_without_total = led_filter.split(" ")[0] if led_filter else None
    filters = led_filter_without_total.split("+") if led_filter_without_total else []
    exposure_duration_setting_key = (
        f"{led_filter_without_total} Exposure"
        if led_filter_without_total and len(filters) == 1
        else "Exposure"
    )
    illumination_setting_key = (
        f"{led_filter_without_total} Preset Intensity"
        if led_filter_without_total and len(filters) == 1
        else "Preset Intensity"
    )

    feature_names = _build_feature_names(filters)
    features = [
        feature
        for feature_name in feature_names
        if (feature := _build_feature(feature_name, plate_data))
    ]

    # Used to capture extra keys for matching LED filters only.
    filter_numbers = re.findall(r"\d+", led_filter) if led_filter else []
    filter_number_regex = f".*({'|'.join(filter_numbers)})+.*" if filter_numbers else ""
    measurement = Measurement(
        type_=MeasurementType.OPTICAL_IMAGING,
        identifier=random_uuid_str(),
        location_identifier=location_id,
        well_plate_identifier=well_plate,
        sample_identifier=f"{well_plate}_{location_id}",
        detection_type=constants.DETECTION_TYPE,
        device_type=constants.DEVICE_TYPE,
        led_filter=led_filter,
        exposure_duration_setting=plate_data.get(float, exposure_duration_setting_key),
        illumination_setting=plate_data.get(float, illumination_setting_key),
        illumination_setting_unit=UNITLESS,
        processed_data=ProcessedData(
            identifier=random_uuid_str(),
            features=features,
            data_processing_document=plate_data.get_custom_keys(
                {
                    "Masked",
                    # Preset Size for LED filter belongs here.
                    rf".*{filter_number_regex}.*Preset Size.*",
                }
            ),
        ),
        device_control_custom_info=plate_data.get_custom_keys(
            {
                f"Max Sensor Value{filter_number_regex}",
                f"Average Sensor Value{filter_number_regex}",
                f"{filter_number_regex}Calibrated Exposure (b|B)y Mabtech",
                f"{filter_number_regex}Preset AOI",
                f"{filter_number_regex}Preset Emphasis",
                f"{filter_number_regex}Preset Contrast",
                f"{filter_number_regex}Preset Name",
                f"{filter_number_regex}Preset Brightness",
                "Analyte Secreting Population",
            }
        ),
        # Machine ID is read in metadata. The second value filters any remaning columns with 3
        # digit numbers, which filters columns for other LED filters not part of this measurement.
        custom_info=(
            plate_data.get_unread(regex=filter_number_regex)
            if filter_number_regex
            else {}
        )
        | plate_data.get_unread(skip={"Machine ID", r".*\d{3}.*"}),
    )

    if not (measurement.processed_data and measurement.processed_data.features):
        logging.warning(f"no image features identified for {well_plate}")
    return measurement


def _build_feature_names(led_filters: list[str]) -> list[str]:
    feature_names = []
    seen = set()

    for feature in constants.IMAGE_FEATURES:
        if not led_filters:
            formatted_feature = feature.format(filter="", led_number="")
            if formatted_feature not in seen:
                feature_names.append(formatted_feature)
                seen.add(formatted_feature)
            continue
        for led_filter in led_filters:
            led_number = re.search(r"\d+", led_filter)
            if not led_number:
                error_msg = f"Unable to interpret LED number from {led_filter}"
                raise AllotropeParsingError(error_msg)

            formatted_feature = feature.format(
                filter=led_filter, led_number=led_number.group()
            )
            if formatted_feature not in seen:
                feature_names.append(formatted_feature)
                seen.add(formatted_feature)

    return feature_names


def _build_feature(feature: str, plate_data: SeriesData) -> ImageFeature | None:
    value = plate_data.get(float, feature)
    if value is None:
        return None
    return ImageFeature(identifier=random_uuid_str(), feature=feature, result=value)


def create_measurement_group(
    well_data: list[SeriesData], plate_info: SeriesData
) -> MeasurementGroup:
    # These custom info keys should be read into the measurement group
    custom_info_keys = {"Saved Date", "Read Date"}
    for well_row in well_data:
        well_row.mark_read(custom_info_keys)

    measurements = [_create_measurement(well_row) for well_row in well_data]

    return MeasurementGroup(
        measurements=measurements,
        plate_well_count=96,
        measurement_time=well_data[0][str, "Read Date"],
        analyst=plate_info.get(str, "Saved By:"),
        custom_info=well_data[0].get_custom_keys(custom_info_keys),
    )



from allotropy.allotrope.models.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.thermo_fisher_qubit4.thermo_fisher_qubit4_reader import (
    ThermoFisherQubit4Reader,
)
from allotropy.parsers.thermo_fisher_qubit4.thermo_fisher_qubit4_structure import (
    create_measurement_group,
    create_metadata,
)
from allotropy.parsers.utils.pandas import map_rows
from allotropy.parsers.vendor_parser import VendorParser


class ThermoFisherQubit4Parser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Thermo Fisher Scientific Qubit 4"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = ThermoFisherQubit4Reader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        return Data(
            create_metadata(named_file_contents.original_file_path),
            measurement_groups=map_rows(
                ThermoFisherQubit4Reader.read(named_file_contents),
                create_measurement_group,
            ),
        )



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
import re
from re import Match
from typing import ClassVar, TypeAlias

import pandas as pd

from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
)
from allotropy.exceptions import (
    AllotropeConversionError,
    valid_value_or_raise,
)
from allotropy.parsers.bmg_mars import constants
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import assert_not_none


class ReadType(Enum):
    ABSORBANCE = "Absorbance"
    FLUORESCENCE = "Fluorescence"
    LUMINESCENCE = "Luminescence"

    @property
    def measurement_type(self) -> MeasurementType:
        if self is ReadType.ABSORBANCE:
            return MeasurementType.ULTRAVIOLET_ABSORBANCE
        elif self is ReadType.LUMINESCENCE:
            return MeasurementType.LUMINESCENCE
        else:
            return MeasurementType.FLUORESCENCE

    @property
    def device_type(self) -> str:
        return f"{self.detection_type} detector"

    @property
    def detection_type(self) -> str:
        return self.value.lower()


FilterHandler: TypeAlias = Callable[[Match[str]], tuple[float | None, float | None]]


@dataclass(frozen=True)
class Header:
    read_type: ReadType
    wavelength: float | None
    excitation_wavelength: float | None
    user: str
    test_name: str
    date: str
    time: str
    id1: str
    id2: str | None
    id3: str | None
    path: str | None
    test_id: str | None

    FILTER_FORMATS: ClassVar[dict[str, tuple[str, FilterHandler]]] = {
        "Raw Data (Ex/Em)": (
            r"Raw Data \((\d+\.?\d*)/(\d+\.?\d*)\)",
            lambda m: (float(m.group(2)), float(m.group(1))),
        ),
        "Raw Data (Em)": (
            r"Raw Data \((\d+\.?\d*)\)",
            lambda m: (float(m.group(1)), None),
        ),
        "Raw Data (No filter)": (
            r"Raw Data \(No filter\)",
            lambda _: (None, None),
        ),
    }

    @staticmethod
    def create(data: SeriesData, header_content: str) -> Header:
        # Search for read type in header content
        read_types = {
            read_type
            for read_type in ReadType
            if read_type.value.lower() in header_content.lower()
        }
        read_type = valid_value_or_raise("read type", read_types, ReadType)

        # Get wavelengths from RawData line
        raw_data_line = assert_not_none(
            re.search(r"Raw Data \(.*?\)", header_content),
            msg="Raw Data line not found in input file.",
        ).group(0)

        # iterate over filter formats, to parse Ex/Em
        for _, (pattern, handler) in Header.FILTER_FORMATS.items():
            match = re.match(pattern, raw_data_line)
            if match:
                try:
                    wavelength, excitation_wavelength = handler(match)
                    break
                except ValueError as err:
                    msg = f"Failed to convert filters in line: '{raw_data_line}'"
                    raise AllotropeConversionError(msg) from err
        else:
            msg = (
                f"Invalid Raw Data format: '{raw_data_line}'. "
                "Expected/supported formats include: "
                f"{', '.join(Header.FILTER_FORMATS.keys())}"
            )
            raise AllotropeConversionError(msg)

        return Header(
            read_type=read_type,
            wavelength=wavelength,
            excitation_wavelength=excitation_wavelength,
            user=data[str, "USER"],
            path=data.get(str, "PATH"),
            test_id=data.get(str, "TEST ID"),
            test_name=data[str, "TEST NAME"],
            date=data[str, "DATE"],
            time=data[str, "TIME"],
            id1=data[str, "ID1"],
            id2=data.get(str, "ID2"),
            id3=data.get(str, "ID3"),
        )


def create_metadata(header: Header, file_path: str) -> Metadata:
    asm_file_identifier = Path(file_path).with_suffix(".json")
    return Metadata(
        file_name=Path(file_path).name,
        asm_file_identifier=asm_file_identifier.name,
        unc_path=header.path or file_path,
        device_identifier=NOT_APPLICABLE,
        model_number=NOT_APPLICABLE,
        data_system_instance_id=NOT_APPLICABLE,
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        software_name=constants.SOFTWARE_NAME,
    )


def _create_measurement(
    row_name: str,
    col_name: str,
    value: float,
    header: Header,
) -> Measurement:
    well_location = f"{row_name}{col_name}"
    return Measurement(
        type_=header.read_type.measurement_type,
        identifier=random_uuid_str(),
        sample_identifier=f"{header.id1} {well_location}",
        location_identifier=well_location,
        well_plate_identifier=header.id1,
        device_type=header.read_type.device_type,
        detection_type=header.read_type.detection_type,
        detector_wavelength_setting=header.wavelength,
        excitation_wavelength_setting=header.excitation_wavelength,
        absorbance=value if header.read_type is ReadType.ABSORBANCE else None,
        fluorescence=value if header.read_type is ReadType.FLUORESCENCE else None,
        luminescence=value if header.read_type is ReadType.LUMINESCENCE else None,
    )


def create_measurement_groups(
    data: pd.DataFrame,
    header: Header,
) -> list[MeasurementGroup]:
    return [
        MeasurementGroup(
            measurement_time=f"{header.date} {header.time}",
            plate_well_count=data.size,
            experiment_type=header.test_name,
            experimental_data_identifier=header.id2,
            analyst=header.user,
            measurements=[
                _create_measurement(str(row_name), str(col_name), value, header)
            ],
        )
        for row_name, row in data.iterrows()
        for col_name, value in row.items()
        if not pd.isnull(value)
    ]



from allotropy.allotrope.models.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.thermo_fisher_nanodrop_8000.nanodrop_8000_reader import (
    Nanodrop8000Reader,
)
from allotropy.parsers.thermo_fisher_nanodrop_8000.nanodrop_8000_structure import (
    create_measurement_group,
    create_metadata,
    SpectroscopyRow,
)
from allotropy.parsers.utils.pandas import map_rows
from allotropy.parsers.vendor_parser import VendorParser


class Nanodrop8000Parser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Thermo Fisher Scientific NanoDrop 8000"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = Nanodrop8000Reader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = Nanodrop8000Reader(named_file_contents)
        rows = map_rows(reader.data, SpectroscopyRow.create)

        return Data(
            create_metadata(named_file_contents.original_file_path),
            measurement_groups=[create_measurement_group(row) for row in rows],
            # NOTE: in current implementation, calculated data is reported at global level for some reason.
            # TODO(nstender): should we move this inside of measurements?
            calculated_data=[item for row in rows for item in row.calculated_data],
        )



from allotropy.allotrope.models.adm.cell_counting.rec._2024._09.cell_counting import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.cell_counting.rec._2024._09.cell_counting import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.roche_cedex_hires.constants import DISPLAY_NAME
from allotropy.parsers.roche_cedex_hires.roche_cedex_hires_reader import (
    RocheCedexHiResReader,
)
from allotropy.parsers.roche_cedex_hires.roche_cedex_hires_structure import (
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.utils.pandas import map_rows
from allotropy.parsers.vendor_parser import VendorParser


class RocheCedexHiResParser(VendorParser[Data, Model]):
    DISPLAY_NAME = DISPLAY_NAME
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = RocheCedexHiResReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = RocheCedexHiResReader(named_file_contents)
        return Data(
            create_metadata(reader.header, named_file_contents.original_file_path),
            map_rows(reader.data, create_measurement_groups),
        )



# generated by datamodel-codegen:
#   filename:  light-obscuration.schema.json
#   timestamp: 2024-06-04T03:30:35+00:00

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCountsPerMilliliter,
    TQuantityValueMicrometer,
    TQuantityValueMilliliter,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TBooleanValue,
    TDateTimeStampValue,
    TIntValue,
    TStringValue,
)


@dataclass(kw_only=True)
class DataSourceDocumentItem:
    data_source_identifier: TStringValue
    data_source_feature: TStringValue


@dataclass(kw_only=True)
class TDataSourceAggregateDocument:
    data_source_document: list[DataSourceDocumentItem]


@dataclass(kw_only=True)
class SampleDocument:
    sample_identifier: TStringValue


@dataclass(kw_only=True)
class Manifest:
    vocabulary: list[str]
    json_schemas: list[str]
    field_id: str | None = None
    field_type: str | None = None
    shapes: list[str] | None = None


@dataclass(kw_only=True)
class DataSystemDocument:
    data_system_instance_identifier: TStringValue | None = None
    file_name: TStringValue | None = None
    UNC_path: TStringValue | None = None
    software_name: TStringValue | None = None
    software_version: TStringValue | None = None
    ASM_converter_name: TStringValue | None = None
    ASM_converter_version: TStringValue | None = None


@dataclass(kw_only=True)
class DeviceDocumentItem:
    detector_identifier: TStringValue
    detector_model_number: TStringValue
    field_index: int | None = None


@dataclass(kw_only=True)
class DeviceSystemDocument:
    asset_management_identifier: TStringValue | None = None
    description: Any | None = None
    brand_name: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    device_identifier: TStringValue | None = None
    model_number: TStringValue | None = None
    equipment_serial_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    device_document: list[DeviceDocumentItem] | None = None


@dataclass(kw_only=True)
class DeviceControlDocumentItem:
    flush_volume_setting: TQuantityValueMilliliter
    detector_view_volume: TQuantityValueMilliliter
    repetition_setting: TIntValue
    sample_volume_setting: TQuantityValueMilliliter
    field_index: int | None = None


@dataclass(kw_only=True)
class DeviceControlAggregateDocument:
    device_control_document: list[DeviceControlDocumentItem] | None = None


@dataclass(kw_only=True)
class CalculatedDataDocumentItem:
    calculated_data_name: TStringValue
    calculated_result: TQuantityValueCountsPerMilliliter | TQuantityValueMicrometer | TQuantityValueUnitless
    data_source_aggregate_document: TDataSourceAggregateDocument | None = None
    calculated_data_identifier: TStringValue | None = None
    calculation_description: TStringValue | None = None


@dataclass(kw_only=True)
class TCalculatedDataAggregateDocument:
    calculated_data_document: list[CalculatedDataDocumentItem]


@dataclass(kw_only=True)
class DataProcessingDocument:
    dilution_factor_setting: TQuantityValueUnitless | None = None
    data_processing_omission_setting: TBooleanValue | None = None


@dataclass(kw_only=True)
class DistributionItem:
    particle_size: TQuantityValueMicrometer
    cumulative_count: TQuantityValueUnitless
    cumulative_particle_density: TQuantityValueCountsPerMilliliter
    distribution_identifier: TStringValue | None = None
    differential_particle_density: TQuantityValueCountsPerMilliliter | None = None
    differential_count: TQuantityValueUnitless | None = None


@dataclass(kw_only=True)
class DistributionDocumentItem:
    distribution: list[DistributionItem]
    field_index: int | None = None


@dataclass(kw_only=True)
class DistributionAggregateDocument:
    distribution_document: list[DistributionDocumentItem]


@dataclass(kw_only=True)
class ProcessedDataDocumentItem:
    data_processing_document: DataProcessingDocument | None = None
    distribution_aggregate_document: DistributionAggregateDocument | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class ProcessedDataAggregateDocument:
    processed_data_document: list[ProcessedDataDocumentItem]


@dataclass(kw_only=True)
class MeasurementDocumentItem:
    measurement_identifier: TStringValue
    measurement_time: TDateTimeStampValue
    device_control_aggregate_document: DeviceControlAggregateDocument | None = None
    sample_document: SampleDocument | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None


@dataclass(kw_only=True)
class MeasurementAggregateDocument:
    analyst: TStringValue
    submitter: TStringValue | None = None
    measurement_document: list[MeasurementDocumentItem] | None = None


@dataclass(kw_only=True)
class LightObscurationDocumentItem:
    measurement_aggregate_document: MeasurementAggregateDocument


@dataclass(kw_only=True)
class LightObscurationAggregateDocument:
    light_obscuration_document: list[LightObscurationDocumentItem]
    data_system_document: DataSystemDocument | None = None
    device_system_document: DeviceSystemDocument | None = None
    calculated_data_aggregate_document: TCalculatedDataAggregateDocument | None = None


@dataclass(kw_only=True)
class Model:
    field_asm_manifest: Manifest | str
    light_obscuration_aggregate_document: LightObscurationAggregateDocument | None = (
        None
    )



from allotropy.calcdocs.extractor import Element, Extractor
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.structure import (
    WellItem,
)


class AppbioQuantstudioDAExtractor(Extractor[WellItem]):
    @classmethod
    def to_element(cls, well_item: WellItem) -> Element:
        return Element(
            data={
                "uuid": well_item.uuid,
                "identifier": well_item.identifier,
                "target_dna_description": well_item.target_dna_description,
                "sample_identifier": well_item.sample_identifier,
                "reporter_dye_setting": well_item.reporter_dye_setting,
                "well_location_identifier": well_item.well_location_identifier,
                "quencher_dye_setting": well_item.quencher_dye_setting,
                "sample_role_type": well_item.sample_role_type,
                "cycle_threshold_value_setting": well_item.result.cycle_threshold_value_setting,
                "cycle_threshold_result": well_item.result.cycle_threshold_result,
                "automatic_cycle_threshold_enabled_setting": well_item.result.automatic_cycle_threshold_enabled_setting,
                "automatic_baseline_determination_enabled_setting": well_item.result.automatic_baseline_determination_enabled_setting,
                "normalized_reporter_result": well_item.result.normalized_reporter_result,
                "baseline_corrected_reporter_result": well_item.result.baseline_corrected_reporter_result,
                "baseline_determination_start_cycle_setting": well_item.result.baseline_determination_start_cycle_setting,
                "baseline_determination_end_cycle_setting": well_item.result.baseline_determination_end_cycle_setting,
                "genotyping_determination_result": well_item.result.genotyping_determination_result,
                "genotyping_determination_method_setting": well_item.result.genotyping_determination_method_setting,
                "quantity": well_item.result.quantity,
                "quantity_mean": well_item.result.quantity_mean,
                "quantity_sd": well_item.result.quantity_sd,
                "ct_mean": well_item.result.ct_mean,
                "eq_ct_mean": well_item.result.eq_ct_mean,
                "adj_eq_ct_mean": well_item.result.adj_eq_ct_mean,
                "ct_sd": well_item.result.ct_sd,
                "ct_se": well_item.result.ct_se,
                "delta_ct_mean": well_item.result.delta_ct_mean,
                "delta_ct_se": well_item.result.delta_ct_se,
                "delta_ct_sd": well_item.result.delta_ct_sd,
                "delta_delta_ct": well_item.result.delta_delta_ct,
                "rq": well_item.result.rq,
                "rq_min": well_item.result.rq_min,
                "rq_max": well_item.result.rq_max,
                "rn_mean": well_item.result.rn_mean,
                "rn_sd": well_item.result.rn_sd,
                "y_intercept": well_item.result.y_intercept,
                "r_squared": well_item.result.r_squared,
                "slope": well_item.result.slope,
                "efficiency": well_item.result.efficiency,
                "amp_score": well_item.result.amp_score,
                "cq_conf": well_item.result.cq_conf,
            }
        )



from dataclasses import dataclass

import pandas as pd

from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_reader import (
    DesignQuantstudioReader,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.structure import (
    Well,
    WellItem,
    WellList,
)
from allotropy.parsers.utils.pandas import (
    assert_df_column,
)


@dataclass
class SimplePrimaryAnalysisWellItem(WellItem):
    @classmethod
    def get_amplification_data_sheet(
        cls, reader: DesignQuantstudioReader
    ) -> pd.DataFrame | None:
        return reader.get_non_empty_sheet_or_none("Amplification Data")


@dataclass
class SimplePrimaryAnalysisWell(Well):
    @classmethod
    def get_well_item_class(cls) -> type[WellItem]:
        return SimplePrimaryAnalysisWellItem


@dataclass(frozen=True)
class SimplePrimaryAnalysisWellList(WellList):
    @classmethod
    def get_well_class(cls) -> type[Well]:
        return SimplePrimaryAnalysisWell

    @classmethod
    def get_well_result_data(cls, reader: DesignQuantstudioReader) -> pd.DataFrame:
        results_data = super(  # noqa: UP008
            SimplePrimaryAnalysisWellList, cls
        ).get_well_result_data(reader)

        if "Well" not in results_data:
            pos = assert_df_column(results_data, "Well Position")
            pos_to_id = {p: i for i, p in enumerate(pos.unique(), start=1)}
            results_data["Well"] = [pos_to_id[pos] for pos in pos]

        if "Target" not in results_data:
            results_data["Target"] = "N/A"

        return results_data



from __future__ import annotations

from io import BytesIO
from re import search
from zipfile import ZipFile

from allotropy.parsers.utils.values import assert_not_none


def fix_zip(data: BytesIO) -> BytesIO:
    # ZipFile can fail if there are excess trailing bytes. This function detects the end of zip's central directory and,
    # if found, truncates the data after the end of the zip file.
    content = data.read()
    pos = content.rfind(
        b"\x50\x4b\x05\x06"
    )  # reverse find: this string of bytes is the end of the zip's central directory.
    if pos > 0:
        data.seek(
            pos + 22
        )  # End bytes size is 22, see https://pkwaredownloads.blob.core.windows.net/pkware-general/Documentation/APPNOTE-6.3.0.TXT
        data.truncate()
        data.seek(0)

    return data


class ZipHandler:
    def __init__(self, data: BytesIO):
        self.zip_file = ZipFile(fix_zip(data))
        self.name_list = self.zip_file.namelist()

    def get_inner_path_or_none(self, pattern: str) -> str | None:
        for file_name in self.name_list:
            if search(pattern, file_name):
                return file_name
        return None

    def get_inner_path(self, pattern: str) -> str:
        return assert_not_none(
            self.get_inner_path_or_none(pattern),
            msg=f"Unable to find file that match pattern {pattern}",
        )

    def get_file(self, inner_path: str) -> BytesIO:
        return BytesIO(self.zip_file.read(inner_path))

    def get_zip(self, inner_path: str) -> ZipHandler:
        return ZipHandler(self.get_file(inner_path))

    def get_file_from_pattern(self, pattern: str) -> BytesIO:
        return self.get_file(self.get_inner_path(pattern))

    def get_zip_from_pattern(self, pattern: str) -> ZipHandler:
        return self.get_zip(self.get_inner_path(pattern))



from enum import Enum

from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    TransmittedLightSetting,
)

MULTIPLATE_FILE_ERROR = "Only a single plate per file can be processed at this time. Please refer to Gen5 documentation for how to generate single plate exports from multi-plate experiments"
NO_PLATE_DATA_ERROR = "No plate data found in file."
UNSUPPORTED_READ_TYPE_ERROR = "Only imaging results can be processed at this time."

HEADER_PREFIXES = frozenset(
    {
        "Software Version",
        "Experiment File Path:",
        "Protocol File Path:",
        "Plate Number",
        "Date",
        "Time",
        "Reader Type:",
        "Reader Serial Number:",
        "Reading Type",
    },
)

DEFAULT_SOFTWARE_NAME = "Gen5 Image"
DEVICE_TYPE = "Imager"
DETECTION_TYPE = "optical-imaging"
METADATA_ONLY = "METADATA ONLY"

AUTOFOCUS_STRINGS = frozenset(
    {
        "Laser autofocus",
        "Autofocus with optional scan",
        "Autofocus without optional scan",
    }
)

DETECTOR_DISTANCE_REGEX = r"Fixed focal height at bottom elevation plus (-?[\d\.]+) \w."
CHANNEL_HEADER_REGEX = r"\tChannel \d+:  (?P<fluorescent_tag>.+) (?P<excitation_wavelength>\d{3}),(?P<detector_wavelength>\d{3})"
FILENAME_REGEX = r"^\d{6}_\d{6}_(?P<plate_identifier>.[^_]*)_.*\.txt$"
SETTINGS_SECTION_REGEX = r"^\tChannel|^\tColor Camera"

TRANSMITTED_LIGHT_MAP = {
    "Brightfield": TransmittedLightSetting.brightfield,
    "Phase Contrast": TransmittedLightSetting.phase_contrast,
    "Transmitted Light": TransmittedLightSetting.brightfield,
    "Reflected Light": TransmittedLightSetting.brightfield,
    "Color Bright Field": TransmittedLightSetting.brightfield,
    "Bright Field": TransmittedLightSetting.brightfield,
    "Bright Field:High Contrast": TransmittedLightSetting.brightfield,
}


class DetectionType(str, Enum):
    SINGLE_IMAGE = "Image Single Image"
    MONTAGE = "Image Montage"
    Z_STACK = "Image Z-Stack"


class ReadType(str, Enum):
    ENDPOINT = "Endpoint"
    KINETIC = "Kinetic"
    AREASCAN = "Area Scan"
    SPECTRAL = "Spectral"
    IMAGE = "Image"



from __future__ import annotations

import numpy as np
import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import (
    df_to_series_data,
    parse_header_row,
    read_multisheet_excel,
    SeriesData,
)
from allotropy.parsers.utils.values import assert_not_none


class MabtechApexReader:
    SUPPORTED_EXTENSIONS = "xlsx"

    plate_info: SeriesData
    data: pd.DataFrame

    @staticmethod
    def create(named_file_contents: NamedFileContents) -> MabtechApexReader:
        raw_contents = read_multisheet_excel(named_file_contents.contents)
        return MabtechApexReader(raw_contents)

    def __init__(self, raw_contents: dict[str, pd.DataFrame]) -> None:
        contents = {
            str(name): df.replace(np.nan, None) for name, df in raw_contents.items()
        }

        self.plate_info = self._get_plate_info(contents)
        self.data = self._get_data(contents)

    def _get_plate_info(self, contents: dict[str, pd.DataFrame]) -> SeriesData:
        sheet = (
            assert_not_none(
                contents.get("Plate Information"),
                msg="Unable to find 'Plate Information' sheet.",
            )
            .dropna(axis=1, how="all")
            .T
        )
        return df_to_series_data(parse_header_row(sheet))

    def _get_data(self, contents: dict[str, pd.DataFrame]) -> pd.DataFrame:
        sheet = assert_not_none(
            contents.get("Plate Database"), msg="Unable to find 'Plate Database' sheet."
        )

        return sheet.dropna(axis="columns", how="all")



from allotropy.allotrope.models.adm.plate_reader.rec._2024._06.plate_reader import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import CsvReader, read_to_lines
from allotropy.parsers.perkin_elmer_envision.perkin_elmer_envision_structure import (
    create_calculated_data,
    create_measurement_groups,
    create_metadata,
    Data as StructureData,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class PerkinElmerEnvisionParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "PerkinElmer Envision"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = "csv"
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        lines = read_to_lines(named_file_contents)
        reader = CsvReader(lines)
        data = StructureData.create(reader)

        return Data(
            create_metadata(
                data.software, data.instrument, named_file_contents.original_file_path
            ),
            create_measurement_groups(data),
            create_calculated_data(data.plate_list, data.labels.get_read_type()),
        )



from typing import ClassVar

from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_calculated_documents import (
    iter_primary_analysis_calc_docs,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_reader import (
    DesignQuantstudioReader,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_views import (
    SampleView,
    TargetView,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.constants import (
    ExperimentType,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.creator import (
    Creator,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.structure import (
    Data,
    Header,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.primary_analysis.structure import (
    PrimaryAnalysisWellList,
)


class PrimaryAnalysisCreator(Creator):
    PLUGIN_REGEX: ClassVar[str] = r"^Primary Analysis v\d+\.\d+\.\d+$"
    EXPECTED_SHEETS: ClassVar[list[str]] = [
        "Results",
        "Amplification Data",
        "Multicomponent",
        "Replicate Group Result",
    ]

    @classmethod
    def create(cls, reader: DesignQuantstudioReader) -> Data:
        header = Header.create(reader.header)
        wells = PrimaryAnalysisWellList.create(reader, header)
        well_items = wells.get_well_items()

        return Data(
            header,
            wells,
            experiment_type=ExperimentType.primary_analysis_experiment,
            calculated_documents=list(
                iter_primary_analysis_calc_docs(
                    view_st_data=SampleView(sub_view=TargetView()).apply(well_items),
                )
            ),
            reference_target=None,
            reference_sample=None,
        )



import json
from pathlib import Path
import re
from typing import Any

import numpy as np
import pandas as pd

from allotropy.json_to_csv.mapper_config import (
    DatasetConfig,
    JoinTransformConfig,
    MapperConfig,
    PivotTransformConfig,
)


def _apply_join(
    datasets: dict[str, pd.DataFrame], transform_config: JoinTransformConfig
) -> dict[str, pd.DataFrame]:
    if transform_config.dataset_1 not in datasets:
        msg = (
            f"Invalid join transform, missing dataset_1: {transform_config.dataset_1}."
        )
        raise ValueError(msg)
    if transform_config.join_key_1 not in datasets[transform_config.dataset_1]:
        msg = f"Invalid join transform, dataset_1 ({transform_config.dataset_1}) is missing column for join_key_1: {transform_config.join_key_1}."
        raise ValueError(msg)
    if transform_config.dataset_2 not in datasets:
        msg = (
            f"Invalid join transform, missing dataset_2: {transform_config.dataset_2}."
        )
        raise ValueError(msg)
    if transform_config.join_key_2 not in datasets[transform_config.dataset_2]:
        msg = f"Invalid join transform, dataset_2 ({transform_config.dataset_2}) is missing column for join_key_2: {transform_config.join_key_2}."
        raise ValueError(msg)

    # Copy the dataframe to join, and rename the join key column so that it matches the first dataset.
    to_join = (
        datasets[transform_config.dataset_2]
        .copy(deep=True)
        .rename(columns={transform_config.join_key_2: transform_config.join_key_1})
    )

    # Join with left outer join
    join = datasets[transform_config.dataset_1].merge(
        to_join, on=transform_config.join_key_1, how="left", indicator=True
    )
    join = join[join["_merge"] != "right_only"].drop(columns=["_merge"])

    # Replace the original dataset.
    datasets[transform_config.dataset_1] = join
    return datasets


def _apply_pivot(
    df: pd.DataFrame, transform_config: PivotTransformConfig, config: DatasetConfig
) -> pd.DataFrame:
    # Get the set of non-pivoted columns
    value_column_name = config.path_to_config[transform_config.value_path].name
    label_column_name = config.path_to_config[transform_config.label_path].name
    other_columns = [
        column
        for column in df.columns
        if column not in {value_column_name, label_column_name}
    ]

    # Rename pivot value column and drop the label column
    df, new_column_names = _rename_column(df, value_column_name)
    config.replace_column_names(value_column_name, new_column_names)
    df = df.drop(columns=label_column_name)

    # Group by non-pivot columns, and compress each into a single row.
    rows: list[list[Any]] = []
    for group_values, sub_df in df.groupby(list(other_columns), dropna=False):
        rows.append([])
        for column in sub_df:
            value = {value for value in sub_df[column].unique() if not pd.isna(value)}
            if len(value) > 1:
                msg = f"Multiple unique values for column '{column}' in pivot operation for group: {group_values}: {value}."
                raise ValueError(msg)
            rows[-1].append(next(iter(value)) if value else np.nan)

    return pd.DataFrame(rows, columns=df.columns)


def _map_dataset(
    data: dict[str, Any], config: DatasetConfig, current_path: Path | None = None
) -> pd.DataFrame:
    current_path = current_path or Path()
    # Map simple values from the current level
    single_values = {}
    for key, value in data.items():
        path = Path(current_path, key)
        column_config = config.get_column_config(str(path))
        if not column_config:
            continue
        if not isinstance(value, dict | list):
            single_values[column_config.name] = [value]
    df = pd.DataFrame(single_values)

    # Map nested dictionaries and lists, combining them with the base dataset.
    for key, value in data.items():
        path = Path(current_path, key)
        path_df: pd.DataFrame = pd.DataFrame()

        if isinstance(value, dict):
            path_df = _map_dataset(value, config, path)
        elif isinstance(value, list):
            path_df = pd.concat(
                [_map_dataset(item, config, path) for item in value],
                ignore_index=True,
            )

        for transform in config.path_to_transform.get(str(path), []):
            if isinstance(transform, PivotTransformConfig):
                if not isinstance(value, list):
                    msg = f"Invalid target for pivot transform: '{path}', target must be a list."
                    raise ValueError(msg)
                path_df = _apply_pivot(path_df, transform, config)
            else:  # should not be possible
                msg = f"Invalid transform type in dataset config: {transform.type_}"
                raise ValueError(msg)

        df = (
            df
            if path_df.empty
            else (path_df if df.empty else df.merge(path_df, how="cross"))
        )

    return df


def _rename_column(
    df: pd.DataFrame, column_name: str
) -> tuple[pd.DataFrame, list[str]]:
    if column_name not in df:
        return df, []

    # Get column names to be used for rename.
    labels = re.findall(r"\$([^\$]*)\$", column_name)
    label_values = df.loc[:, labels]

    # Get unique combinations of column names.
    unique_values = [tuple(t) for _, t in label_values.drop_duplicates().iterrows()]
    new_column_values = {
        unique_value: [np.nan] * df.shape[0] for unique_value in unique_values
    }
    # Map the unique labels to the corresponding column value.
    for index, row in df.iterrows():
        label_tuple = tuple(row.loc[labels])
        new_column_values[label_tuple][int(str(index))] = row.loc[column_name]

    # Drop the old column
    insert_index = df.columns.get_loc(column_name)
    df = df.drop(columns=[column_name])

    # Inject the new columns
    new_column_names: list[str] = []
    for unique_tuple in unique_values:
        new_column_name = column_name
        for label, value in zip(labels, unique_tuple, strict=True):
            new_column_name = new_column_name.replace(f"${label}$", value)
        new_column_names.append(new_column_name)
        df.insert(insert_index, new_column_name, new_column_values[unique_tuple])
        insert_index += 1

    return df, new_column_names


def map_dataset(data: dict[str, Any], config: DatasetConfig) -> pd.DataFrame:
    df = _map_dataset(data, config)

    # Check that required columns are populated.
    required_columns = {column.name for column in config.columns if column.required}
    if missing_columns := required_columns - set(df.columns):
        msg = f"Mapped dataset is missing required columns: {missing_columns}"
        raise ValueError(msg)

    # Put columns in the order of the config
    if config.columns:
        df = df[[column for column in config.column_names if column in df]]

    # Rename columns with substitution labels
    for column in config.columns:
        if column.has_labels:
            df, _ = _rename_column(df, column.name)

    # Drop columns that should not be included
    columns_to_drop = [
        column.name
        for column in config.columns
        if not column.include and column.name in df
    ]
    df = df.drop(columns=columns_to_drop)

    return df


def _convert_to_json_blob(df: pd.DataFrame, config: DatasetConfig) -> dict[str, Any]:
    try:
        blob = json.loads(df.to_json())
    except ValueError as e:
        msg = f"Unable to convert dataset {config.name} to json blob with error: {e}"
        raise ValueError(msg) from e

    if not isinstance(blob, dict):
        msg = f"Invalid candidate for dataset to metadata json conversion for dataset {config.name} result is not a dictionary: {blob}"
        raise ValueError(msg)

    cleaned: dict[str, Any] = {}
    for key, value in blob.items():
        if not isinstance(value, dict):
            msg = f"Invalid object: {value} at key {key} when converting dataframe to json."
            raise ValueError(msg)
        if len(value) != 1:
            msg = f"Invalid candidate for dataset to metadata json conversion for key {key}: {value}. Must have a single value per key."
            raise ValueError(msg)
        cleaned[key] = value[next(iter(value))]
    return cleaned


def json_to_csv(
    data: dict[str, Any], config: MapperConfig
) -> dict[str, pd.DataFrame | dict[str, Any]]:
    datasets = {
        name: map_dataset(data, dataset_config)
        for name, dataset_config in config.datasets.items()
    }

    # Apply transforms across multiple datasets (e.g. joins)
    for transform in config.transforms:
        if isinstance(transform, JoinTransformConfig):
            _apply_join(datasets, transform)

    # Remove datasets that should not be included.
    datasets = {
        name: dataset
        for name, dataset in datasets.items()
        if config.datasets[name].include
    }

    # Convert metadata datasets to json
    result: dict[str, pd.DataFrame | dict[str, Any]] = {
        name: _convert_to_json_blob(dataset, config.datasets[name])
        if config.datasets[name].is_metadata
        else dataset
        for name, dataset in datasets.items()
    }
    return result



from __future__ import annotations

from allotropy.allotrope.models.adm.plate_reader.rec._2024._06.plate_reader import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.unchained_labs_lunatic.unchained_labs_lunatic_reader import (
    UnchainedLabsLunaticReader,
)
from allotropy.parsers.unchained_labs_lunatic.unchained_labs_lunatic_structure import (
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.vendor_parser import VendorParser


class UnchainedLabsLunaticParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Unchained Labs Lunatic"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = UnchainedLabsLunaticReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = UnchainedLabsLunaticReader(named_file_contents)
        measurement_groups, calculated_data = create_measurement_groups(
            reader.header, reader.data
        )
        return Data(
            create_metadata(reader.header, named_file_contents.original_file_path),
            measurement_groups=measurement_groups,
            calculated_data=calculated_data,
        )



from __future__ import annotations

from dataclasses import dataclass

from allotropy.calcdocs.view import ViewData


@dataclass(frozen=True)
class CalculatedDataConfig:
    name: str
    value: str
    view_data: ViewData
    source_configs: list[CalculatedDataConfig | MeasurementConfig]


@dataclass(frozen=True)
class MeasurementConfig:
    name: str
    value: str



from dataclasses import dataclass
from typing import Any

from allotropy.allotrope.converter import add_custom_information_document
from allotropy.allotrope.models.adm.solution_analyzer.rec._2024._09.solution_analyzer import (
    AnalyteAggregateDocument,
    AnalyteDocument,
    CalculatedDataAggregateDocument,
    CalculatedDataDocumentItem,
    DataProcessingDocument,
    DataSourceAggregateDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    DistributionAggregateDocument,
    DistributionDocumentItem,
    ErrorAggregateDocument,
    ErrorDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocument,
    Model,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
    SolutionAnalyzerAggregateDocument,
    SolutionAnalyzerDocumentItem,
    TQuantityValueModel,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCell,
    TQuantityValueCountsPerMilliliter,
    TQuantityValueDegreeCelsius,
    TQuantityValueGramPerLiter,
    TQuantityValueMicrometer,
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueMilliliter,
    TQuantityValueMilliliterPerLiter,
    TQuantityValueMillimeterOfMercury,
    TQuantityValueMillimolePerLiter,
    TQuantityValueMillionCellsPerMilliliter,
    TQuantityValueMilliOsmolesPerKilogram,
    TQuantityValuePercent,
    TQuantityValuePH,
    TQuantityValueUnitless,
)
from allotropy.allotrope.schema_mappers.schema_mapper import SchemaMapper
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.utils.values import quantity_or_none


@dataclass(frozen=True)
class Analyte:
    name: str
    value: float
    unit: str

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, Analyte):
            return False

        return self.name < other.name


@dataclass(frozen=True)
class DistributionDocument:
    particle_size: float
    cumulative_count: float
    cumulative_particle_density: float
    differential_particle_density: float
    differential_count: float
    distribution_identifier: str


@dataclass(frozen=True)
class Error:
    error: str
    feature: str | None = None


@dataclass(frozen=True)
class DataProcessing:
    cell_type_processing_method: str | None = None
    cell_density_dilution_factor: float | None = None
    dilution_factor_setting: float | None = None
    data_processing_omission_setting: bool | None = None


@dataclass(frozen=True)
class Measurement:
    # Measurement metadata
    identifier: str
    measurement_time: str
    sample_identifier: str
    description: str | None = None
    batch_identifier: str | None = None
    detection_type: str | None = None

    # Measurements
    absorbance: float | None = None
    po2: float | None = None
    pco2: float | None = None
    carbon_dioxide_saturation: float | None = None
    oxygen_saturation: float | None = None
    ph: float | None = None
    temperature: float | None = None
    osmolality: float | None = None
    viability: float | None = None
    total_cell_density: float | None = None
    viable_cell_density: float | None = None
    average_live_cell_diameter: float | None = None
    total_cell_count: float | None = None
    viable_cell_count: float | None = None
    analytes: list[Analyte] | None = None
    data_processing: DataProcessing | None = None
    distribution_documents: list[DistributionDocument] | None = None

    # Errors
    errors: list[Error] | None = None


@dataclass(frozen=True)
class MeasurementGroup:
    measurements: list[Measurement]
    analyst: str | None
    data_processing_time: str | None
    errors: list[Error] | None = None


@dataclass(frozen=True)
class Metadata:
    device_type: str
    asm_file_identifier: str
    data_system_instance_identifier: str
    unc_path: str
    device_identifier: str | None = None
    model_number: str | None = None
    software_name: str | None = None
    detection_type: str | None = None
    software_version: str | None = None
    equipment_serial_number: str | None = None
    product_manufacturer: str | None = None
    brand_name: str | None = None
    file_name: str | None = None
    analyst: str | None = None
    measurement_time: str | None = None
    analytical_method_identifier: str | None = None
    method_version: str | None = None
    experimental_data_identifier: str | None = None
    flush_volume_setting: float | None = None
    detector_view_volume: float | None = None
    repetition_setting: int | None = None
    sample_volume_setting: float | None = None


@dataclass(frozen=True)
class DataSource:
    identifier: str
    feature: str


@dataclass(frozen=True)
class CalculatedDataItem:
    identifier: str
    name: str
    value: float
    unit: str
    data_sources: list[DataSource]


@dataclass(frozen=True)
class Data:
    metadata: Metadata
    measurement_groups: list[MeasurementGroup]
    calculated_data: list[CalculatedDataItem] | None = None


class Mapper(SchemaMapper[Data, Model]):
    MANIFEST = "http://purl.allotrope.org/manifests/solution-analyzer/REC/2024/09/solution-analyzer.manifest"

    def map_model(self, data: Data) -> Model:
        return Model(
            solution_analyzer_aggregate_document=add_custom_information_document(
                SolutionAnalyzerAggregateDocument(
                    device_system_document=add_custom_information_document(
                        DeviceSystemDocument(
                            model_number=data.metadata.model_number,
                            equipment_serial_number=data.metadata.equipment_serial_number,
                            device_identifier=data.metadata.device_identifier,
                            product_manufacturer=data.metadata.product_manufacturer,
                        ),
                        None,
                    ),
                    data_system_document=add_custom_information_document(
                        DataSystemDocument(
                            ASM_file_identifier=data.metadata.asm_file_identifier,
                            data_system_instance_identifier=data.metadata.data_system_instance_identifier,
                            file_name=data.metadata.file_name,
                            UNC_path=data.metadata.unc_path,
                            software_name=data.metadata.software_name,
                            software_version=data.metadata.software_version,
                            ASM_converter_name=self.converter_name,
                            ASM_converter_version=ASM_CONVERTER_VERSION,
                        ),
                        None,
                    ),
                    solution_analyzer_document=[
                        self._get_technique_document(measurement_group, data.metadata)
                        for measurement_group in data.measurement_groups
                    ],
                    calculated_data_aggregate_document=self._get_calculated_data_aggregate_document(
                        data.calculated_data
                    ),
                ),
                None,
            ),
            field_asm_manifest=self.MANIFEST,
        )

    def _get_technique_document(
        self, measurement_group: MeasurementGroup, metadata: Metadata
    ) -> SolutionAnalyzerDocumentItem:
        return SolutionAnalyzerDocumentItem(
            analyst=measurement_group.analyst,
            measurement_aggregate_document=add_custom_information_document(
                MeasurementAggregateDocument(
                    data_processing_time=self.get_date_time(
                        measurement_group.data_processing_time
                    )
                    if measurement_group.data_processing_time
                    else None,
                    measurement_document=[
                        self._get_measurement_document_item(measurement, metadata)
                        for measurement in measurement_group.measurements
                    ],
                    error_aggregate_document=self._get_error_aggregate_document(
                        measurement_group.errors
                    ),
                ),
                None,
            ),
        )

    def _get_measurement_document_item(
        self, measurement: Measurement, metadata: Metadata
    ) -> MeasurementDocument:
        return add_custom_information_document(
            MeasurementDocument(
                measurement_identifier=measurement.identifier,
                measurement_time=self.get_date_time(measurement.measurement_time),
                sample_document=self._get_sample_document(measurement),
                device_control_aggregate_document=DeviceControlAggregateDocument(
                    device_control_document=[
                        add_custom_information_document(
                            DeviceControlDocumentItem(
                                device_type=metadata.device_type,
                                detection_type=measurement.detection_type,
                                flush_volume_setting=quantity_or_none(
                                    TQuantityValueMilliliter,
                                    metadata.flush_volume_setting,
                                ),
                                detector_view_volume=quantity_or_none(
                                    TQuantityValueMilliliter,
                                    metadata.detector_view_volume,
                                ),
                                repetition_setting=metadata.repetition_setting,
                                sample_volume_setting=quantity_or_none(
                                    TQuantityValueMilliliter,
                                    metadata.sample_volume_setting,
                                ),
                            ),
                            None,
                        ),
                    ]
                ),
                analyte_aggregate_document=AnalyteAggregateDocument(
                    analyte_document=[
                        self._create_analyte_document(analyte)
                        for analyte in measurement.analytes
                    ]
                )
                if measurement.analytes
                else None,
                processed_data_aggregate_document=self._create_processed_data_document(
                    measurement
                ),
                error_aggregate_document=self._get_error_aggregate_document(
                    measurement.errors
                ),
                absorbance=quantity_or_none(
                    TQuantityValueMilliAbsorbanceUnit, measurement.absorbance
                ),
                pO2=quantity_or_none(
                    TQuantityValueMillimeterOfMercury, measurement.po2
                ),
                pCO2=quantity_or_none(
                    TQuantityValueMillimeterOfMercury, measurement.pco2
                ),
                carbon_dioxide_saturation=quantity_or_none(
                    TQuantityValuePercent, measurement.carbon_dioxide_saturation
                ),
                oxygen_saturation=quantity_or_none(
                    TQuantityValuePercent, measurement.oxygen_saturation
                ),
                pH=quantity_or_none(TQuantityValuePH, measurement.ph),
                temperature=quantity_or_none(
                    TQuantityValueDegreeCelsius, measurement.temperature
                ),
                osmolality=quantity_or_none(
                    TQuantityValueMilliOsmolesPerKilogram, measurement.osmolality
                ),
            ),
            None,
        )

    def _get_sample_document(self, measurement: Measurement) -> SampleDocument:
        return add_custom_information_document(
            SampleDocument(
                sample_identifier=measurement.sample_identifier,
                batch_identifier=measurement.batch_identifier,
                description=measurement.description,
            ),
            None,
        )

    def _create_analyte_document(self, analyte: Analyte) -> AnalyteDocument:
        if analyte.unit == "g/L":
            return AnalyteDocument(
                analyte_name=analyte.name,
                mass_concentration=TQuantityValueGramPerLiter(value=analyte.value),
            )
        elif analyte.unit == "mL/L":
            return AnalyteDocument(
                analyte_name=analyte.name,
                volume_concentration=TQuantityValueMilliliterPerLiter(
                    value=analyte.value
                ),
            )
        elif analyte.unit == "mmol/L":
            return AnalyteDocument(
                analyte_name=analyte.name,
                molar_concentration=TQuantityValueMillimolePerLiter(
                    value=analyte.value
                ),
            )
        elif analyte.unit == "U/L":
            if analyte.name == "ldh":
                return AnalyteDocument(
                    analyte_name=analyte.name,
                    molar_concentration=TQuantityValueMillimolePerLiter(
                        value=analyte.value * 0.0167
                        if analyte.value > 0
                        else analyte.value
                    ),
                )
            else:
                msg = f"Invalid unit for {analyte.name}: {analyte.unit}"
                raise AllotropeConversionError(msg)

        msg = f"Invalid unit for analyte: {analyte.unit}, value values are: g/L, mL/L, mmol/L"
        raise AllotropeConversionError(msg)

    def _create_processed_data_document(
        self, measurement: Measurement
    ) -> ProcessedDataAggregateDocument | None:
        processed_data_document = add_custom_information_document(
            ProcessedDataDocumentItem(
                viability__cell_counter_=quantity_or_none(
                    TQuantityValuePercent, measurement.viability
                ),
                total_cell_density__cell_counter_=quantity_or_none(
                    TQuantityValueMillionCellsPerMilliliter,
                    measurement.total_cell_density,
                ),
                viable_cell_density__cell_counter_=quantity_or_none(
                    TQuantityValueMillionCellsPerMilliliter,
                    measurement.viable_cell_density,
                ),
                average_live_cell_diameter__cell_counter_=quantity_or_none(
                    TQuantityValueMicrometer, measurement.average_live_cell_diameter
                ),
                total_cell_count=quantity_or_none(
                    TQuantityValueCell, measurement.total_cell_count
                ),
                viable_cell_count=quantity_or_none(
                    TQuantityValueCell, measurement.viable_cell_count
                ),
                data_processing_document=DataProcessingDocument(
                    cell_type_processing_method=measurement.data_processing.cell_type_processing_method,
                    cell_density_dilution_factor=quantity_or_none(
                        TQuantityValueUnitless,
                        measurement.data_processing.cell_density_dilution_factor,
                    ),
                    dilution_factor_setting=quantity_or_none(
                        TQuantityValueUnitless,
                        measurement.data_processing.dilution_factor_setting,
                    ),
                    data_processing_omission_setting=measurement.data_processing.data_processing_omission_setting,
                )
                if measurement.data_processing
                else None,
                distribution_aggregate_document=DistributionAggregateDocument(
                    distribution_document=[
                        DistributionDocumentItem(
                            distribution_identifier=distribution.distribution_identifier,
                            particle_size=TQuantityValueMicrometer(
                                value=distribution.particle_size
                            ),
                            cumulative_count=TQuantityValueUnitless(
                                value=distribution.cumulative_count
                            ),
                            cumulative_particle_density=TQuantityValueCountsPerMilliliter(
                                value=distribution.cumulative_particle_density
                            ),
                            differential_particle_density=TQuantityValueCountsPerMilliliter(
                                value=distribution.differential_particle_density
                            ),
                            differential_count=TQuantityValueUnitless(
                                value=distribution.differential_count
                            ),
                        )
                        for distribution in measurement.distribution_documents
                    ]
                )
                if measurement.distribution_documents
                else None,
            ),
            None,
        )

        if all(value is None for value in processed_data_document.__dict__.values()):
            return None

        return ProcessedDataAggregateDocument(
            processed_data_document=[processed_data_document]
        )

    def _get_error_aggregate_document(
        self, errors: list[Error] | None
    ) -> ErrorAggregateDocument | None:
        if not errors:
            return None

        return ErrorAggregateDocument(
            error_document=[
                ErrorDocumentItem(error=error.error, error_feature=error.feature)
                for error in errors
            ]
        )

    def _get_calculated_data_aggregate_document(
        self, calculated_data_items: list[CalculatedDataItem] | None
    ) -> CalculatedDataAggregateDocument | None:
        if not calculated_data_items:
            return None

        return CalculatedDataAggregateDocument(
            calculated_data_document=[
                CalculatedDataDocumentItem(
                    calculated_data_identifier=calculated_data_item.identifier,
                    calculated_data_name=calculated_data_item.name,
                    calculated_result=TQuantityValueModel(
                        value=calculated_data_item.value,
                        unit=calculated_data_item.unit,
                    ),
                    data_source_aggregate_document=DataSourceAggregateDocument(
                        data_source_document=[
                            DataSourceDocumentItem(
                                data_source_identifier=item.identifier,
                                data_source_feature=item.feature,
                            )
                            for item in calculated_data_item.data_sources
                        ]
                    ),
                )
                for calculated_data_item in calculated_data_items
            ]
        )



from allotropy.allotrope.models.adm.plate_reader.rec._2024._06.plate_reader import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    Data,
    Mapper,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.agilent_gen5.agilent_gen5_reader import AgilentGen5Reader
from allotropy.parsers.agilent_gen5.agilent_gen5_structure import (
    create_kinetic_results,
    create_metadata,
    create_results,
    get_identifiers,
    get_kinetic_measurements,
    get_results_section,
    get_temperature,
    HeaderData,
    KineticData,
    ReadData,
)
from allotropy.parsers.agilent_gen5.constants import (
    NO_MEASUREMENTS_ERROR,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class AgilentGen5Parser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Agilent Gen5"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = AgilentGen5Reader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = AgilentGen5Reader(named_file_contents)

        if (results_section := get_results_section(reader)) is None:
            raise AllotropeConversionError(NO_MEASUREMENTS_ERROR)

        header_data = HeaderData.create(
            reader.header_data, named_file_contents.original_file_path
        )
        read_data = ReadData.create(reader.sections["Procedure Details"])
        kinetic_data = KineticData.create(reader.sections["Procedure Details"])

        sample_identifiers = get_identifiers(reader.sections.get("Layout"))
        actual_temperature = get_temperature(reader.sections.get("Actual Temperature"))
        kinetic_measurements, kinetic_elapsed_time = get_kinetic_measurements(
            reader.sections.get("Time")
        ) or ({}, [])

        if kinetic_data and not (kinetic_measurements and kinetic_elapsed_time):
            msg = "Kinetic data is present in the file but no kinetic measurements data is found."
            raise AllotropeConversionError(msg)

        # in case of single kinetic section, each well will contain one measurement document with the kinetic data cube
        # otherwise each well will contain multiple measurement documents
        if kinetic_data:
            measurement_groups, calculated_data = create_kinetic_results(
                results_section,
                header_data,
                read_data,
                sample_identifiers,
                actual_temperature,
                kinetic_data,
                kinetic_measurements,
                kinetic_elapsed_time,
            )
        else:
            measurement_groups, calculated_data = create_results(
                results_section,
                header_data,
                read_data,
                sample_identifiers,
                actual_temperature,
            )

        if not measurement_groups:
            raise AllotropeConversionError(NO_MEASUREMENTS_ERROR)

        return Data(
            metadata=create_metadata(header_data),
            measurement_groups=measurement_groups,
            calculated_data=calculated_data,
        )



from __future__ import annotations

from collections.abc import Callable, Iterable
from enum import Enum
import os
import re
from typing import Any, Literal, overload, TypeVar
import unicodedata
import warnings

import pandas as pd
from pandas._typing import FilePath, ReadCsvBuffer

from allotropy.allotrope.models.shared.definitions.definitions import (
    InvalidJsonFloat,
)
from allotropy.exceptions import (
    AllotropeConversionError,
    AllotropeParsingError,
)
from allotropy.parsers.utils.iterables import get_first_not_none
from allotropy.parsers.utils.values import (
    assert_is_type,
    assert_not_none,
    str_to_bool,
    try_float_or_none,
)
from allotropy.types import IOType

MapType = TypeVar("MapType")


def map_rows(
    data_frame: pd.DataFrame, func: Callable[[SeriesData], MapType]
) -> list[MapType]:
    def run_with_data(series: pd.Series[str]) -> MapType:
        return func(SeriesData(series))

    # pandas can't find a matching overload for this, but it works and returns the correct type...
    return list(data_frame.apply(run_with_data, axis="columns"))  # type: ignore[call-overload]


def rm_df_columns(data: pd.DataFrame, pattern: str) -> pd.DataFrame:
    return data.drop(
        columns=[column for column in data.columns if re.match(pattern, column)]
    )


def set_columns(data: pd.DataFrame, column_names: Iterable[str]) -> None:
    cols = list(column_names)
    if data.shape[1] != len(cols):
        msg = f"Invalid format - mismatch between # of columns ({data.shape[1]}) and # of labels ({len(cols)}), column labels: {cols}."
        raise AllotropeConversionError(msg)
    data.columns = pd.Index(cols)


def df_to_series(df: pd.DataFrame, index: int | None = None) -> pd.Series[Any]:
    df = df.dropna(how="all")
    n_rows, _ = df.shape
    if index is None and n_rows != 1:
        msg = "Unable to convert DataFrame to series: data has more than 1 row and no index was provided."
        raise AllotropeConversionError(msg)
    index = index or 0
    if index == -1:
        index = n_rows - 1
    if index >= n_rows:
        msg = f"Index {index} is greater than the number of rows in dataframe {n_rows}."
        raise AllotropeConversionError(msg)
    return pd.Series(df.iloc[index], index=df.columns)


def df_to_series_data(df: pd.DataFrame, index: int | None = None) -> SeriesData:
    df.columns = df.columns.astype(str).str.strip()
    return SeriesData(df_to_series(df, index))


def assert_df_column(df: pd.DataFrame, column: str) -> pd.Series[Any]:
    df_column = df.get(column)
    if df_column is None:
        msg = f"Unable to find column '{column}'."
        raise AllotropeConversionError(msg)
    return pd.Series(df_column)


def assert_not_empty_df(df: pd.DataFrame, msg: str) -> pd.DataFrame:
    if df.empty:
        raise AllotropeConversionError(msg)
    return df


def assert_value_from_df(df: pd.DataFrame, key: str) -> Any:
    try:
        return df[key]
    except KeyError as e:
        msg = f"Unable to find key '{key}' in dataframe headers: {df.columns.tolist()}"
        raise AllotropeConversionError(msg) from e


def parse_header_row(df: pd.DataFrame) -> pd.DataFrame:
    # Set the first row of a dataframe as the columns. This is useful when the dataframe has already been
    # parsed (e.g. when splitting a dataframe into two parts), and so the header/index_col arguments cannot
    # be used to set the columns/index at read time.
    df.columns = pd.Index(
        assert_not_empty_df(df, "Cannot set parse header row for empty dataframe.")
        .astype(str)
        .iloc[0]
    )
    df.columns = df.astype(str).columns.str.strip()
    return df[1:]


def split_header_and_data(
    df: pd.DataFrame, should_split_on_row: Callable[[pd.Series[Any]], bool]
) -> tuple[pd.DataFrame, pd.DataFrame]:
    header, data = split_dataframe(df, should_split_on_row)
    if data is None:
        msg = f"Unable to split header and data from dataframe: {df}"
        raise AllotropeConversionError(msg)

    return header, data


def split_dataframe(
    df: pd.DataFrame,
    should_split_on_row: Callable[[pd.Series[Any]], bool],
    *,
    include_split_row: bool = False,
) -> tuple[pd.DataFrame, pd.DataFrame | None]:
    for idx, row in df.iterrows():
        if should_split_on_row(row):
            head_end = int(str(idx))
            tail_start = head_end if include_split_row else head_end + 1
            return df[:head_end], df[tail_start:]

    return df, None


def _normalize_columns(df: pd.DataFrame) -> pd.DataFrame:
    return df.rename(
        columns=lambda col: (
            unicodedata.normalize("NFKC", col) if isinstance(col, str) else col
        )
    )


def read_csv(
    # types for filepath_or_buffer match those in pd.read_csv()
    filepath_or_buffer: FilePath | ReadCsvBuffer[bytes] | ReadCsvBuffer[str],
    **kwargs: Any,
) -> pd.DataFrame:
    """Wrap pd.read_csv() and raise AllotropeParsingError for failures.

    pd.read_csv() can return a DataFrame or TextFileReader. The latter is intentionally not supported.
    """
    try:
        df_or_reader = pd.read_csv(filepath_or_buffer, **kwargs)
    except Exception as e:
        msg = f"Error calling pd.read_csv(): {e}"
        raise AllotropeParsingError(msg) from e
    return _normalize_columns(
        assert_is_type(
            df_or_reader,
            pd.DataFrame,
            "pd.read_csv() returned a TextFileReader, which is not supported.",
        )
    )


def read_excel(
    # io is untyped in pd.read_excel(), but this seems reasonable.
    io: str | IOType,
    **kwargs: Any,
) -> pd.DataFrame:
    """Wrap pd.read_excel() and raise AllotropeParsingError for failures.

    pd.read_excel() can return a DataFrame or a dictionary of DataFrames. The latter is intentionally not supported.
    """
    try:
        df_or_dict = pd.read_excel(io, **kwargs)
    except Exception as e:
        msg = f"Error calling pd.read_excel(): {e}"
        raise AllotropeParsingError(msg) from e
    return _normalize_columns(
        assert_is_type(df_or_dict, pd.DataFrame, "Expected a single-sheet Excel file.")
    )


def read_multisheet_excel(
    io: str | IOType,
    **kwargs: Any,
) -> dict[str, pd.DataFrame]:
    try:
        df_or_dict = pd.read_excel(io, sheet_name=None, **kwargs)
    except Exception as e:
        msg = f"Error calling pd.read_excel(): {e}"
        raise AllotropeParsingError(msg) from e
    sheets: dict[str, pd.DataFrame] = assert_is_type(
        df_or_dict, dict, "Expected a multi-sheet Excel file."
    )
    return {
        sheet_name: _normalize_columns(
            assert_is_type(df, pd.DataFrame, "Expected all sheets to yield dataframes.")
        )
        for sheet_name, df in sheets.items()
    }


T = TypeVar("T", bool, float, int, str)
# Need to use this instead of type[T] to get mypy to realize primitive can be called to return T
Type_ = Callable[..., T]
KeyOrKeys = Iterable[str]
TypeAndKey = tuple[Type_[T], KeyOrKeys]
TypeAndKeyAndMsg = tuple[Type_[T], KeyOrKeys, str]
ValidateRaw = Callable[[Any], bool] | None


class SeriesData:
    class ValidateRawMode(Enum):
        # Return None for key is raw value is None or np.isna
        NOT_NAN = "NOT_NAN"

    NOT_NAN = ValidateRawMode.NOT_NAN

    @staticmethod
    def _validate_raw(v: Any, mode: ValidateRawMode | None) -> Any:
        if mode is SeriesData.ValidateRawMode.NOT_NAN:
            return None if (v is None or pd.isna(v)) else v
        return v

    def __init__(self, series: pd.Series[Any]) -> None:
        self.series = series
        self.read_keys: set[str] = set()
        self.errored = False

    def __del__(self) -> None:
        if self.errored:
            return
        # NOTE: this will be turned on by default when all callers have been updated to pass the warning.
        if unread_keys := set(self.series.index.to_list()) - self.read_keys:
            if os.getenv("WARN_UNUSED_KEYS"):
                warnings.warn(
                    f"SeriesData went out of scope without reading all keys, unread: {sorted(unread_keys)}.",
                    stacklevel=2,
                )

    def _get_custom_key(self, key: str) -> float | str | None:
        if (float_value := self.get(float, key)) is not None:
            return float_value
        return self.get(str, key)

    def _get_matching_keys(self, key_or_keys: str | set[str]) -> set[str]:
        return {
            matched
            for regex_key in (
                key_or_keys if isinstance(key_or_keys, set) else {key_or_keys}
            )
            for matched in [
                k
                for k in self.series.index
                if k == regex_key or re.fullmatch(regex_key, k)
            ]
        }

    def get_custom_keys(
        self, key_or_keys: str | set[str]
    ) -> dict[str, float | str | None]:
        return {
            key: self._get_custom_key(key)
            for key in self._get_matching_keys(key_or_keys)
        }

    def mark_read(self, key_or_keys: str | set[str]) -> None:
        self.read_keys |= self._get_matching_keys(key_or_keys)

    def get_unread(
        self, regex: str | None = None, skip: set[str] | None = None
    ) -> dict[str, float | str | None]:
        skip = self._get_matching_keys(skip) if skip else set()
        # Mark explicitly skipped keys as "read". This not only covers the check below, but removes
        # them from the destructor warning.
        self.read_keys |= skip
        matching_keys = (
            self._get_matching_keys(regex)
            if regex
            else set(self.series.index.to_list())
        )
        return self.get_custom_keys(matching_keys - self.read_keys)

    def has_key(self, key: str) -> bool:
        return key in self.series

    def __getitem__(self, type_and_key: TypeAndKey[T] | TypeAndKeyAndMsg[T]) -> T:
        """
        Get a value of the specified type with the specified key, raising an error if the
        key is not found, or if the value cannot be converted to the type.
        If a third argument is provided, it is an error message to provide if they key is not found.

        value: float = series_data[float, key]
        value: str = series_data[str, key, f"Failed to find {key} in my data"]

        Parameters:
        type (str, int, float, bool): The datatype to return.
        key (str | Iterable[str]): The key (or iterable of keys) to use to lookup.
        msg (str | None): The message to give as an error if lookup or conversion fails.

        Returns:
        type: A value of the type provided.

        Raises
        AllotropeConversionError: If the lookup or conversion to type fails.
        """
        if len(type_and_key) == 2:
            type_, key = type_and_key
            msg = None
        elif len(type_and_key) == 3:
            type_, key, msg = type_and_key
        try:
            return assert_not_none(self.get(type_, key), str(key), msg=msg)
        except Exception:
            self.errored = True
            raise

    # This overload tells typing that if default is "None" then get might return None
    @overload
    def get(
        self,
        type_: Type_[T],
        key: KeyOrKeys,
        default: Literal[None] = None,
        validate: ValidateRawMode | None = None,
    ) -> T | None:
        ...

    # This overload tells typing that if default matches T, get will return T
    @overload
    def get(
        self,
        type_: Type_[float],
        key: KeyOrKeys,
        default: InvalidJsonFloat,
        validate: ValidateRawMode | None = None,
    ) -> float | InvalidJsonFloat:
        ...

    # This overload tells typing that if default matches T, get will return T
    @overload
    def get(
        self,
        type_: Type_[T],
        key: KeyOrKeys,
        default: T,
        validate: ValidateRawMode | None = None,
    ) -> T:
        ...

    def get(
        self,
        type_: Type_[T],
        key: KeyOrKeys,
        default: T | InvalidJsonFloat | None = None,
        validate: ValidateRawMode | None = None,
    ) -> T | InvalidJsonFloat | None:
        """
        Get a value of the specified type with the specified key, returning a default value if the
        key is not found, or if the value cannot be converted to the type.
        If validate is provided, it will be used to validate the raw value from the series, returning None
        if the value fails validation.

        Parameters:
        type (str, int, float, bool): The datatype to return.
        key (str | Iterable[str]): The key (or iterable of keys) to use to lookup.
        default (type | InvalidJsonFloat | None): The value to return if lookup or conversion fails (default=None).
        validate (ValidateRawMode): The method to use for validating raw value. Defaults to (value is not None).

        Returns:
        type: A value of the type provided or default value.
        """
        if not isinstance(key, str):
            return get_first_not_none(
                lambda k: self.get(type_, k, validate=validate), key
            )
        self.read_keys.add(key)
        raw_value = self._validate_raw(self.series.get(key), validate)
        try:
            # bool needs special handling to convert
            if type_ is bool:
                raw_value = (
                    None
                    if raw_value is None
                    else ("true" if str_to_bool(str(raw_value)) else "")
                )
            if type_ is float and isinstance(raw_value, str) and "%" in raw_value:
                raw_value = raw_value.strip("%")
            convert = try_float_or_none if type_ is float else type_
            # mypy can't figure out that try_float_or_none will only be used when type_ is float.
            value = None if raw_value is None else convert(raw_value)  # type: ignore[operator]
        except ValueError:
            value = None
        return default if value is None else value



from __future__ import annotations

from pathlib import Path

import pandas as pd

from allotropy.allotrope.models.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    ContainerType,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    NaN,
)
from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Data,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
)
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.thermo_fisher_qubit_flex import constants
from allotropy.parsers.utils.pandas import df_to_series_data, map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


def create_measurement_group(data: SeriesData) -> MeasurementGroup:
    return MeasurementGroup(
        measurement_time=data[str, "Test Date"],
        experiment_type=data.get(str, "Assay Name"),
        measurements=[
            Measurement(
                type_=MeasurementType.FLUORESCENCE,
                identifier=random_uuid_str(),
                fluorescence=data[float, "Sample RFU"],
                batch_identifier=data.get(str, "Run ID"),
                sample_identifier=data[str, ["Sample ID", "Sample Name"]],
                location_identifier=data.get(str, "Well"),
                well_plate_identifier=data.get(str, "Plate Barcode"),
                sample_volume_setting=data.get(
                    float, "Sample Volume (uL)", validate=SeriesData.NOT_NAN
                ),
                excitation_wavelength_setting=data.get(str, "Excitation"),
                dilution_factor_setting=data.get(
                    float, "Dilution Factor", validate=SeriesData.NOT_NAN
                ),
                original_sample_concentration=data.get(
                    float, "Original Sample Conc.", NaN
                ),
                original_sample_concentration_unit=data.get(
                    str, "Original sample conc. units"
                ),
                sample_custom_info={
                    "last read standards": data.get(str, "Last Read Standards"),
                    "selected samples": data.get(int, "Selected Samples"),
                    "qubit tube concentration": {
                        "value": data.get(float, "Qubit Tube Conc.", NaN),
                        "unit": data.get(str, "Qubit tube conc. units", UNITLESS),
                    },
                    "standard 1 concentration": {
                        "value": data.get(
                            float, "Std 1 RFU", validate=SeriesData.NOT_NAN
                        ),
                        "unit": "RFU",
                    },
                    "standard 2 concentration": {
                        "value": data.get(
                            float, "Std 2 RFU", validate=SeriesData.NOT_NAN
                        ),
                        "unit": "RFU",
                    },
                    "standard 3 concentration": {
                        "value": data.get(
                            float, "Std 3 RFU", validate=SeriesData.NOT_NAN
                        ),
                        "unit": "RFU",
                    },
                },
                device_control_custom_info={
                    "operating minimum": {
                        "value": data.get(str, "Extended Low Range"),
                        "unit": "ng/µL",
                    },
                    "operating maximum": {
                        "value": data.get(str, "Extended High Range"),
                        "unit": "ng/µL",
                    },
                    "operating range": {
                        "value": data.get(
                            str, "Core Range", validate=SeriesData.NOT_NAN
                        ),
                        "unit": "ng/µL",
                    },
                },
                custom_info=data.get_unread(skip={"Software Version"}),
            )
        ],
    )


def create_data(df: pd.DataFrame, file_path: str) -> Data:
    header = df_to_series_data(df.head(1))
    # We read header info from a row in the table, so we don't need to read all keys from this SeriesData
    software_version = header.get(str, "Software Version")
    header.get_unread()
    return Data(
        metadata=Metadata(
            file_name=Path(file_path).name,
            unc_path=file_path,
            device_identifier=NOT_APPLICABLE,
            model_number=constants.MODEL_NUMBER,
            software_name=constants.SOFTWARE_NAME,
            software_version=software_version,
            product_manufacturer=constants.PRODUCT_MANUFACTURER,
            brand_name=constants.BRAND_NAME,
            device_type=constants.DEVICE_TYPE,
            container_type=ContainerType.tube,
        ),
        measurement_groups=map_rows(df, create_measurement_group),
    )



from enum import Enum

MULTIPLATE_FILE_ERROR = "Only a single plate per file can be processed at this time. Please refer to Gen5 documentation for how to generate single plate exports from multi-plate experiments"
NO_PLATE_DATA_ERROR = "No plate data found in file."
UNSUPPORTED_READ_TYPE_ERROR = (
    "Only Endpoint measurements can be processed at this time."
)
NO_MEASUREMENTS_ERROR = "Invalid data - the file contains invalid or missing measurement data. Unable to construct ASM."

HEADER_PREFIXES = frozenset(
    {
        "Software Version",
        "Experiment File Path:",
        "Protocol File Path:",
        "Plate Number",
        "Date",
        "Time",
        "Reader Type:",
        "Reader Serial Number:",
        "Reading Type",
    },
)


EMISSION_KEY = "Emission"
EXCITATION_KEY = "Excitation"
GAIN_KEY = "Gain"
OPTICS_KEY = "Optics"
MEASUREMENTS_DATA_POINT_KEY = "Measurements/Data Point"
MIRROR_KEY = "Mirror"
PATHLENGTH_CORRECTION_KEY = "Pathlength Correction"
READ_HEIGHT_KEY = "Read Height"
READ_SPEED_KEY = "Read Speed"
WAVELENGTHS_KEY = "Wavelengths"

DEFAULT_SOFTWARE_NAME = "Gen5"
DEVICE_TYPE = "plate reader"

FILENAME_REGEX = r"^\d{6}_\d{6}_(?P<plate_identifier>.[^_]*)_.*\.txt$"

NAN_EMISSION_EXCITATION = ["Full light", "Plug", "Hole"]


class ReadMode(str, Enum):
    ABSORBANCE = "Absorbance"
    FLUORESCENCE = "Fluorescence"
    LUMINESCENCE = "Luminescence"
    ALPHALISA = "Alpha"


UNSUPPORTED_READ_MODE_ERROR = (
    f"Read mode not found; expected to find one of {sorted(ReadMode._member_names_)}."
)
ELAPSED_TIME = "elapsed time"

READ_DATA_MEASUREMENT_ERROR = "No read mode found for measurement {}"
ALPHALISA_FLUORESCENCE_FOUND = (
    "Both ALPHALISA and FLUORESCENCE read modes detected, which is not supported."
)

DATA_SOURCE_FEATURE_VALUES = {
    ReadMode.ABSORBANCE: "absorption profile data cube",
    ReadMode.FLUORESCENCE: "fluorescence emission profile data cube",
    ReadMode.LUMINESCENCE: "luminescence profile data cube",
}

READ_MODE_UNITS = {
    ReadMode.ABSORBANCE: "mAU",
    ReadMode.FLUORESCENCE: "RFU",
    ReadMode.LUMINESCENCE: "RLU",
}
SECONDS = "s"


class ReadType(str, Enum):
    ENDPOINT = "Endpoint"
    AREASCAN = "Area Scan"
    SPECTRAL = "Spectral"



""" Constants file for Thermo Fisher Qubit Flex Parser"""

# Instrument constants
SOFTWARE_NAME = "Qubit Flex software"
MODEL_NUMBER = "Qubit Flex"
PRODUCT_MANUFACTURER = "Thermo Fisher Scientific"
BRAND_NAME = "Qubit"
DEVICE_TYPE = "fluorescence detector"
DISPLAY_NAME = "Thermo Fisher Scientific Qubit Flex"

# Error constants
VALUE_ERROR = "Unable to find value for column"



#!/usr/bin/env python3
from collections import defaultdict
from datetime import datetime
import os
from pathlib import Path
import subprocess

import click
from dateutil import tz
import semantic_version  # type: ignore

from allotropy.__about__ import __version__
from allotropy.allotrope.schema_parser.path_util import ALLOTROPY_DIR, ROOT_DIR

SECTION_TO_PREFIX = {
    "feat": "Added",
    "fix": "Fixed",
    "refactor": "Changed",
    "deprecate": "Deprecated",
    "remove": "Removed",
    "security": "Security",
}


def _get_changes() -> dict[str, list[str]]:
    p = subprocess.run(
        ["git", "log", "--oneline"], capture_output=True, text=True, check=True
    )
    changes = defaultdict(list)
    for line in p.stdout.split("\n"):
        parts = line.split(" ")
        if len(parts) < 3 or not parts[1].endswith(":"):
            continue
        prefix = parts[1].strip(":").lower()
        if prefix == "release":
            break
        if prefix not in SECTION_TO_PREFIX:
            continue
        changes[prefix].append(" ".join(parts[2:]))
    return dict(changes)


def _get_new_section() -> str:
    changes = _get_changes()
    body = ""
    for prefix, section in SECTION_TO_PREFIX.items():
        if prefix not in changes:
            continue
        body += f"\n### {section}\n\n"
        for change in changes[prefix]:
            body += f"- {change}\n"

    return body


def _update_changelog(version: str) -> str:
    changelog_file = Path(ROOT_DIR, "CHANGELOG.md")
    with open(changelog_file) as f:
        contents = f.readlines()

    body = ""
    with open(changelog_file, "w") as f:
        for line in contents:
            if line.startswith("## ") and not body:
                body = _get_new_section()
                f.write(
                    f"## [{version}] - {datetime.now(tz.gettz('EST')).strftime('%Y-%m-%d')}\n"
                )
                f.write(body)
                f.write("\n")
            f.write(line)

    return body


def _write_version_file(version: str) -> None:
    with open(Path(ALLOTROPY_DIR, "__about__.py"), "w") as f:
        f.write(f'__version__ = "{version}"\n')


def _make_pr(version: str, body: str) -> None:
    print("Making commit...")
    subprocess.run(
        [
            "git",
            "commit",
            "-am",
            f"Update allotropy version to {version}",
        ],
        check=True,
    )
    print("Pushing commit...")
    subprocess.run(["git", "push", "-u", "origin", f"release-v{version}"], check=True)

    print("Tagging branch...")
    subprocess.run(
        ["git", "tag", "-a", f"v{version}", "-m", f"'allotropy v{version}'"], check=True
    )
    subprocess.run(["git", "push", "origin", "tag", f"v{version}"], check=True)

    filename = ".temp_pr_description.txt"
    with open(filename, "w") as f:
        f.write(body)

    try:
        print("Creating PR...")
        subprocess.run(
            [
                "gh",
                "pr",
                "create",
                "--title",
                f"release: Update allotropy version to {version}",
                "--body-file",
                filename,
            ],
            check=True,
        )
        print("Creating release...")
        subprocess.run(
            [
                "gh",
                "release",
                "create",
                f"v{version}",
                "-F",
                filename,
            ],
            check=True,
        )
    finally:
        os.remove(filename)


@click.command()
@click.option(
    "--version", "-v", help="Version to update to, defaults to next patch version"
)
@click.option(
    "--skip_pr", is_flag=True, default=False, help="Whether to make PR in script"
)
def _update_version(
    version: str | None = None, skip_pr: bool = False  # noqa: FBT001, FBT002
) -> None:
    # Check if gh is installed, if not, print hint and exit.
    if not skip_pr:
        try:
            subprocess.run(
                ["gh", "--help"],
                check=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
        except FileNotFoundError:
            print(
                "gh not installed - cannot create PR and release automatically. Try: 'brew install gh'"
            )
            return

    """Update allotropy version."""
    if version:
        semver = semantic_version.Version(version)
    if not version:
        semver = semantic_version.Version(__version__)
        semver.patch += 1

    version = str(semver)

    # Checkout a new branch
    print("Checking out a clean branch from main...")
    subprocess.run(["git", "checkout", "main"], check=True)
    subprocess.run(["git", "pull"], check=True)
    subprocess.run(["git", "checkout", "-b", f"release-v{version}"], check=True)

    print("Updating version file and CHANGELOG...")
    _write_version_file(version)
    body = _update_changelog(version)

    if not skip_pr:
        _make_pr(version, body)


if __name__ == "__main__":
    _update_version()



from dataclasses import dataclass
from typing import Any, TypeVar

from allotropy.allotrope.converter import add_custom_information_document
from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import (
    BaselineCorrectedReporterDataCube,
    CalculatedDataDocumentItem,
    ContainerType,
    DataProcessingDocument,
    DataSourceAggregateDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    ExperimentType,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    MeltingCurveDataCube,
    Model,
    NormalizedReporterDataCube,
    PassiveReferenceDyeDataCube,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    QPCRAggregateDocument,
    QPCRDocumentItem,
    ReporterDyeDataCube,
    SampleDocument,
    TCalculatedDataAggregateDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TNullableQuantityValueUnitless,
    TQuantityValueNumber,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import InvalidJsonFloat
from allotropy.allotrope.schema_mappers.data_cube import DataCube, get_data_cube
from allotropy.allotrope.schema_mappers.schema_mapper import SchemaMapper
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.parsers.utils.values import assert_not_none, quantity_or_none


@dataclass
class DataSource:
    identifier: str
    feature: str


@dataclass
class CalculatedDataItem:
    identifier: str
    name: str
    value: float
    unit: str
    data_sources: list[DataSource]


@dataclass
class CalculatedData:
    items: list[CalculatedDataItem]
    reference_dna_description: str | None = None
    reference_sample_description: str | None = None


@dataclass
class ProcessedData:
    # Settings
    cycle_threshold_value_setting: float
    automatic_cycle_threshold_enabled_setting: bool | None = None
    automatic_baseline_determination_enabled_setting: bool | None = None
    baseline_determination_start_cycle_setting: int | None = None
    baseline_determination_end_cycle_setting: int | None = None
    genotyping_determination_method_setting: float | None = None

    # Results
    genotyping_determination_result: str | None = None
    cycle_threshold_result: float | None = None
    normalized_reporter_result: float | None = None
    baseline_corrected_reporter_result: float | None = None
    normalized_reporter_data_cube: DataCube | None = None
    baseline_corrected_reporter_data_cube: DataCube | None = None

    # Metadata
    comments: str | None = None

    custom_info: dict[str, Any] | None = None


@dataclass
class Measurement:
    # Measurement metadata
    identifier: str
    timestamp: str
    sample_identifier: str
    target_identifier: str
    group_identifier: str | None = None

    # Settings
    pcr_detection_chemistry: str | None = None

    # Optional measurement metadata
    sample_role_type: str | None = None
    well_location_identifier: str | None = None
    well_plate_identifier: str | None = None

    # Optional settings
    total_cycle_number_setting: float | None = None
    reporter_dye_setting: str | None = None
    quencher_dye_setting: str | None = None
    passive_reference_dye_setting: str | None = None

    # Processed data
    processed_data: ProcessedData | None = None
    reporter_dye_data_cube: DataCube | None = None
    passive_reference_dye_data_cube: DataCube | None = None
    melting_curve_data_cube: DataCube | None = None

    # Custom metadata
    custom_info: dict[str, Any] | None = None
    sample_custom_info: dict[str, Any] | None = None
    device_control_custom_info: dict[str, Any] | None = None


@dataclass
class MeasurementGroup:
    measurements: list[Measurement]
    plate_well_count: int | InvalidJsonFloat
    analyst: str | None = None
    experimental_data_identifier: str | None = None


@dataclass
class Metadata:
    device_identifier: str
    device_type: str
    device_serial_number: str
    model_number: str
    file_name: str
    measurement_method_identifier: str
    unc_path: str | None = None
    experiment_type: ExperimentType | None = None
    container_type: ContainerType | None = None
    data_system_instance_identifier: str | None = None
    software_name: str | None = None
    software_version: str | None = None
    product_manufacturer: str | None = None


@dataclass
class Data:
    metadata: Metadata
    measurement_groups: list[MeasurementGroup]
    calculated_data: CalculatedData | None = None


CubeClass = TypeVar("CubeClass")


class Mapper(SchemaMapper[Data, Model]):
    MANIFEST = "http://purl.allotrope.org/manifests/pcr/BENCHLING/2023/09/qpcr.manifest"

    def map_model(self, data: Data) -> Model:
        return Model(
            qPCR_aggregate_document=QPCRAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    device_identifier=data.metadata.device_identifier,
                    model_number=data.metadata.model_number,
                    device_serial_number=data.metadata.device_serial_number,
                    product_manufacturer=data.metadata.product_manufacturer,
                ),
                data_system_document=DataSystemDocument(
                    data_system_instance_identifier=data.metadata.data_system_instance_identifier,
                    file_name=data.metadata.file_name,
                    UNC_path=data.metadata.unc_path,
                    software_name=data.metadata.software_name,
                    software_version=data.metadata.software_version,
                    ASM_converter_name=self.converter_name,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                qPCR_document=[
                    self._get_technique_document(measurement_group, data.metadata)
                    for measurement_group in data.measurement_groups
                ],
                calculated_data_aggregate_document=self._get_calculated_data_aggregate_document(
                    data
                ),
            )
        )

    def _get_technique_document(
        self, measurement_group: MeasurementGroup, metadata: Metadata
    ) -> QPCRDocumentItem:
        return QPCRDocumentItem(
            analyst=measurement_group.analyst,
            measurement_aggregate_document=MeasurementAggregateDocument(
                experimental_data_identifier=measurement_group.experimental_data_identifier,
                experiment_type=metadata.experiment_type,
                container_type=metadata.container_type,
                plate_well_count=TQuantityValueNumber(
                    value=measurement_group.plate_well_count
                ),
                measurement_document=[
                    self._get_measurement_document_item(measurement, metadata)
                    for measurement in measurement_group.measurements
                ],
            ),
        )

    def _get_measurement_document_item(
        self,
        measurement: Measurement,
        metadata: Metadata,
    ) -> MeasurementDocumentItem:
        measurement_doc = MeasurementDocumentItem(
            measurement_identifier=measurement.identifier,
            measurement_time=self.get_date_time(measurement.timestamp),
            target_DNA_description=measurement.target_identifier,
            sample_document=self._get_sample_document(measurement),
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    DeviceControlDocumentItem(
                        device_type=metadata.device_type,
                        measurement_method_identifier=metadata.measurement_method_identifier,
                        total_cycle_number_setting=quantity_or_none(
                            TQuantityValueNumber, measurement.total_cycle_number_setting
                        ),
                        PCR_detection_chemistry=measurement.pcr_detection_chemistry,
                        reporter_dye_setting=measurement.reporter_dye_setting,
                        quencher_dye_setting=measurement.quencher_dye_setting,
                        passive_reference_dye_setting=measurement.passive_reference_dye_setting,
                    )
                ],
            ),
            processed_data_aggregate_document=assert_not_none(
                self._get_processed_data_aggregate_document(measurement.processed_data)
            ),
            reporter_dye_data_cube=get_data_cube(
                measurement.reporter_dye_data_cube, ReporterDyeDataCube
            ),
            passive_reference_dye_data_cube=get_data_cube(
                measurement.passive_reference_dye_data_cube, PassiveReferenceDyeDataCube
            ),
            melting_curve_data_cube=get_data_cube(
                measurement.melting_curve_data_cube, MeltingCurveDataCube
            ),
        )
        return add_custom_information_document(measurement_doc, measurement.custom_info)

    def _get_sample_document(self, measurement: Measurement) -> SampleDocument:
        # TODO(ASM gaps): we believe these values should be added to ASM.
        custom_info_doc = {"group identifier": measurement.group_identifier}
        sample_doc = SampleDocument(
            sample_identifier=measurement.sample_identifier,
            sample_role_type=measurement.sample_role_type,
            well_location_identifier=measurement.well_location_identifier,
            well_plate_identifier=measurement.well_plate_identifier,
        )
        return add_custom_information_document(
            sample_doc, (measurement.sample_custom_info or {}) | custom_info_doc
        )

    def _get_processed_data_aggregate_document(
        self, data: ProcessedData | None
    ) -> ProcessedDataAggregateDocument | None:
        if not data:
            return None
        doc = ProcessedDataDocumentItem(
            data_processing_document=DataProcessingDocument(
                automatic_cycle_threshold_enabled_setting=data.automatic_cycle_threshold_enabled_setting,
                cycle_threshold_value_setting=TQuantityValueUnitless(
                    value=data.cycle_threshold_value_setting,
                ),
                automatic_baseline_determination_enabled_setting=data.automatic_baseline_determination_enabled_setting,
                genotyping_determination_method_setting=quantity_or_none(
                    TQuantityValueUnitless,
                    data.genotyping_determination_method_setting,
                ),
                baseline_determination_start_cycle_setting=quantity_or_none(
                    TQuantityValueNumber,
                    data.baseline_determination_start_cycle_setting,
                ),
                baseline_determination_end_cycle_setting=quantity_or_none(
                    TQuantityValueNumber,
                    data.baseline_determination_end_cycle_setting,
                ),
            ),
            cycle_threshold_result=TNullableQuantityValueUnitless(
                value=data.cycle_threshold_result,
            ),
            normalized_reporter_result=quantity_or_none(
                TQuantityValueUnitless, data.normalized_reporter_result
            ),
            baseline_corrected_reporter_result=quantity_or_none(
                TQuantityValueUnitless,
                data.baseline_corrected_reporter_result,
            ),
            genotyping_determination_result=data.genotyping_determination_result,
            normalized_reporter_data_cube=get_data_cube(
                data.normalized_reporter_data_cube, NormalizedReporterDataCube
            ),
            baseline_corrected_reporter_data_cube=get_data_cube(
                data.baseline_corrected_reporter_data_cube,
                BaselineCorrectedReporterDataCube,
            ),
        )
        return ProcessedDataAggregateDocument(
            processed_data_document=[
                add_custom_information_document(doc, data.custom_info)
            ]
        )

    def _get_calculated_data_aggregate_document(
        self, data: Data
    ) -> TCalculatedDataAggregateDocument | None:
        if not data.calculated_data or not data.calculated_data.items:
            return None

        if data.calculated_data.reference_sample_description:
            data_processing_document = DataProcessingDocument(
                reference_DNA_description=data.calculated_data.reference_dna_description,
                reference_sample_description=data.calculated_data.reference_sample_description,
            )
        else:
            data_processing_document = None

        return TCalculatedDataAggregateDocument(
            calculated_data_document=[
                CalculatedDataDocumentItem(
                    calculated_data_identifier=calc_doc.identifier,
                    data_source_aggregate_document=DataSourceAggregateDocument(
                        data_source_document=[
                            DataSourceDocumentItem(
                                data_source_identifier=data_source.identifier,
                                data_source_feature=data_source.feature,
                            )
                            for data_source in calc_doc.data_sources
                        ],
                    ),
                    data_processing_document=data_processing_document,
                    calculated_data_name=calc_doc.name,
                    calculated_data_description=None,
                    calculated_datum=TQuantityValueUnitless(
                        value=calc_doc.value, unit=calc_doc.unit
                    ),
                )
                for calc_doc in data.calculated_data.items
            ],
        )



from __future__ import annotations

from io import StringIO
import re

import pandas as pd

from allotropy.exceptions import AllotropeConversionError, AllotropeParsingError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import CsvReader, read_to_lines
from allotropy.parsers.luminex_xponent import constants
from allotropy.parsers.utils.pandas import read_csv
from allotropy.parsers.utils.values import assert_not_none, try_float


class LuminexXponentReader:
    SUPPORTED_EXTENSIONS = "csv"

    header_data: pd.DataFrame
    calibration_data: pd.DataFrame
    minimum_assay_bead_count_setting: float
    results_data: dict[str, pd.DataFrame]

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        reader = CsvReader(read_to_lines(named_file_contents))

        self.header_data = self._get_header_data(reader)
        self.calibration_data = self._get_calibration_data(reader)
        self.minimum_assay_bead_count_setting = (
            self._get_minimum_assay_bead_count_setting(reader)
        )
        self.results_data = LuminexXponentReader._get_results(reader)

    @classmethod
    def _get_header_data(cls, reader: CsvReader) -> pd.DataFrame:
        header_lines = list(reader.pop_until(constants.CALIBRATION_BLOCK_HEADER))
        # Header has a weird structure where there rows have varying number of columns, so we need
        # special handling before passing to read_csv.
        n_columns = max(len(line.split(",")) for line in header_lines)
        if n_columns < constants.EXPECTED_HEADER_COLUMNS:
            msg = "Unable to parse header. Not enough data."
            raise AllotropeConversionError(msg)

        return (
            read_csv(
                StringIO("\n".join(header_lines)),
                header=None,
                index_col=0,
                names=range(n_columns),
            )
            .dropna(how="all")
            .T
        )

    @classmethod
    def _get_calibration_data(cls, reader: CsvReader) -> pd.DataFrame:
        reader.drop_until_inclusive(constants.CALIBRATION_BLOCK_HEADER)
        return assert_not_none(
            reader.pop_csv_block_as_df(constants.LUMINEX_EMPTY_PATTERN)
        )

    @classmethod
    def _get_minimum_assay_bead_count_setting(cls, reader: CsvReader) -> float:
        reader.drop_until(match_pat='^"?Samples"?,')
        samples_info = assert_not_none(reader.pop(), msg="Unable to find Samples info.")
        try:
            min_bead_count_setting = samples_info.replace('"', "").split(",")[3]
        except IndexError as e:
            msg = f"Unable to find minimum bead count setting in Samples info: {samples_info}."
            raise AllotropeConversionError(msg) from e

        return try_float(min_bead_count_setting, "minimum bead count setting")

    @staticmethod
    def _get_results(reader: CsvReader) -> dict[str, pd.DataFrame]:
        reader.drop_until_inclusive("Results")
        reader.drop_empty(constants.LUMINEX_EMPTY_PATTERN)
        results: dict[str, pd.DataFrame] = {}
        while reader.current_line_exists() and "-- CRC --" not in (reader.get() or ""):
            result_title_line = assert_not_none(reader.pop())
            match: re.Match[str] | None
            if not (
                match := re.match(constants.TABLE_HEADER_PATTERN, result_title_line)
            ):
                msg = f"Invalid header block start line: {result_title_line}"
                raise AllotropeParsingError(msg)
            result_title = match.groups()[0]
            table_data = assert_not_none(
                reader.pop_csv_block_as_df(
                    empty_pat=constants.LUMINEX_EMPTY_PATTERN, header=0, index_col=0
                )
            ).dropna(how="all", axis="columns")
            results[result_title] = table_data
            reader.drop_empty(constants.LUMINEX_EMPTY_PATTERN)

        return results



from collections.abc import Callable
from typing import Generic, TypeVar

from allotropy.allotrope.models.shared.definitions.definitions import TDateTimeValue

Data = TypeVar("Data")
Model = TypeVar("Model")


class SchemaMapper(Generic[Data, Model]):
    # The manifest of the schema this mapper supports
    MANIFEST: str

    def __init__(
        self, asm_converter_name: str, get_date_time: Callable[[str], TDateTimeValue]
    ) -> None:
        self.converter_name = asm_converter_name
        self.get_date_time = get_date_time

    def map_model(self, data: Data) -> Model:
        raise NotImplementedError



from enum import Enum

DISPLAY_NAME = "Beckman Coulter Biomek"
DEVICE_TYPE = "liquid handler"
PROBE_HEAD_DEVICE_TYPE = "liquid handler probe head"
PRODUCT_MANUFACTURER = "Beckman Coulter"
SOFTWARE_NAME = "BioMek Software"


class TransferStep(str, Enum):
    ASPIRATE = "Aspirate"
    DISPENSE = "Dispense"



import re

from allotropy.allotrope.models.shared.definitions.units import UNITLESS

DETECTION_TYPE = "Absorbance"
DEVICE_TYPE = "plate reader"
MODEL_NUMBER = "Lunatic"
PRODUCT_MANUFACTURER = "Unchained Labs"
SOFTWARE_NAME = "Lunatic and Stunner Analysis"

# Wavelength columns will be "A<3-digit number>" with an optional pathlength specification, e.g. 'A260' or 'A260 (10mm)'
WAVELENGTH_COLUMNS_RE = re.compile(r"^A(\d{3})(?:\s\((\d+)?mm\))?$")
NO_WAVELENGTH_COLUMN_ERROR_MSG = (
    "The file is required to include an absorbance measurement column (e.g. A280)"
)
INCORRECT_WAVELENGTH_COLUMN_FORMAT_ERROR_MSG = (
    msg
) = "The wavelength_column should follow the pattern A### (ex. A260)."
NO_DATE_OR_TIME_ERROR_MSG = "Unable to parse timestamp from input data, expected a 'Date' and 'Time' column in data or a 'Date' in metadata section."
NO_DEVICE_IDENTIFIER_ERROR_MSG = "Unable to parse device identifier, expected an 'Instrument ID' column in data or 'Instrument' in metadata section."
NO_MEASUREMENT_IN_PLATE_ERROR_MSG = (
    "The plate data does not contain absorbance measurement for {}."
)

CALCULATED_DATA_LOOKUP = {
    "A260": [
        {
            "column": "A260 Concentration (ng/ul)",
            "name": "Concentration",
            "feature": "absorbance",
            "unit": "ng/µL",
        },
        {
            "column": "Concentration (ng/ul)",
            "name": "Concentration",
            "feature": "absorbance",
            "unit": "ng/µL",
        },
        {
            "column": "Background (A260)",
            "name": "Background (A260)",
            "feature": "absorbance",
            "unit": "mAU",
        },
        {
            "column": "A260/A230",
            "name": "A260/A230",
            "feature": "absorbance",
            "unit": UNITLESS,
        },
        {
            "column": "A260/A280",
            "name": "A260/A280",
            "feature": "absorbance",
            "unit": UNITLESS,
        },
    ],
    "A280": [
        {
            "column": "Concentration (mg/ml)",
            "name": "Concentration",
            "feature": "absorbance",
            "unit": "mg/mL",
        },
        {
            "column": "Background (A280)",
            "name": "Background (A280)",
            "feature": "absorbance",
            "unit": "mAU",
        },
        {
            "column": "A260/A280",
            "name": "A260/A280",
            "feature": "absorbance",
            "unit": UNITLESS,
        },
    ],
}



from enum import Enum

import numpy as np
import pandas as pd
from pandas import Index

from allotropy.exceptions import AllotropeParsingError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import (
    read_multisheet_excel,
    SeriesData,
    split_dataframe,
)
from allotropy.parsers.utils.values import assert_not_none, try_float_or_none


class SheetNames(Enum):
    PROTOCOL_INFORMATION = "protocol information"
    TABLE_END_POINT = "table end point"
    MICROPLATE_END_POINT = "microplate end point"


class BmgLabtechSmartControlReader:
    SUPPORTED_EXTENSIONS = "xlsx"
    header: SeriesData
    data: pd.DataFrame
    average_of_blank_used: float | None

    def __init__(self, named_file_contents: NamedFileContents):
        raw_contents = read_multisheet_excel(
            named_file_contents.contents,
            header=None,
            engine="calamine",
        )
        contents = {
            name.lower(): df.replace(np.nan, None) for name, df in raw_contents.items()
        }
        for sheet_name in SheetNames:
            if sheet_name.value not in contents:
                msg = f"Sheet '{sheet_name.value}' not found"
                raise AllotropeParsingError(msg)
        self.header = self._get_headers(contents[SheetNames.PROTOCOL_INFORMATION.value])
        self.data = self._get_data(contents[SheetNames.TABLE_END_POINT.value])
        self.average_of_blank_used = self._get_average_of_blank_used(
            contents[SheetNames.MICROPLATE_END_POINT.value]
        )

    def _get_headers(self, data: pd.DataFrame) -> SeriesData:
        headers_df, settings_df = split_dataframe(
            data,
            lambda row: row.astype(str).iloc[0].strip().startswith("Basic settings"),
            include_split_row=True,
        )
        headers_df = assert_not_none(headers_df, "Headers")
        settings_df = assert_not_none(settings_df, "Settings")

        headers_df = headers_df.dropna(how="all")

        split_header = headers_df.iloc[:, 0].str.split(": ", n=1, expand=True)

        headers = pd.DataFrame(
            data=split_header[1].values, index=split_header[0], columns=["Value"]
        )
        settings_df = settings_df.dropna(how="all")
        settings_df = settings_df.set_index(settings_df.columns[0])
        settings_series = settings_df.squeeze()

        # Combine headers and settings into a single Series
        combined_series = pd.concat([headers["Value"], settings_series])
        return SeriesData(combined_series)

    def _get_data(self, data: pd.DataFrame) -> pd.DataFrame:
        _, measurement_data = split_dataframe(
            data,
            lambda row: row.astype(str).iloc[0].strip().startswith("Well"),
            include_split_row=True,
        )
        measurement_data = assert_not_none(measurement_data, "Measurement data")
        measurement_data = measurement_data.transpose()
        measurement_data.columns = Index(measurement_data.iloc[0])
        measurement_data.columns = measurement_data.columns.map(
            lambda x: x.strip() if isinstance(x, str) else x
        )
        measurement_data = measurement_data[1:]
        return measurement_data

    def _get_average_of_blank_used(self, data: pd.DataFrame) -> float | None:
        _, average_of_blank_used_df = split_dataframe(
            data,
            lambda row: row.astype(str).iloc[14].strip().startswith("Blank"),
            include_split_row=True,
        )
        if average_of_blank_used_df is not None:
            average_of_blank_used = average_of_blank_used_df.iloc[0][14]
            return try_float_or_none(average_of_blank_used.strip().split(" ")[-1])
        return None



DISPLAY_NAME = "Thermo Fisher Scientific VISIONlite"
DEVICE_TYPE = "spectrophotometer"
DETECTION_TYPE = "absorbance"
SOFTWARE_NAME = "VISIONlite"
PRODUCT_MANUFACTURER = "Thermo Fisher Scientific"

# Errors
UNSUPPORTED_KINETIC_MEASUREMENTS_ERROR = (
    "Kinetic measurements are not supported at this time."
)

HEADER_COLS = [
    "File Name",
    "Sample Name",
    "Analyst",
    "Measurement Time",
]
SAMPLE_NAME_COLS = ("Sample Name", "Sample name", "sample name", "Well Position")



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
import re
from typing import Any


class TransformType(Enum):
    PIVOT = "PIVOT"
    JOIN = "JOIN"


@dataclass
class TransformConfig:
    type_: TransformType

    @staticmethod
    def create(config_json: dict[str, Any]) -> TransformConfig:
        if "type" not in config_json:
            msg = "Must specify 'type' in transform config"
            raise ValueError(msg)
        try:
            type_ = TransformType(config_json["type"])
        except ValueError as err:
            msg = f"Invalid transform type: {config_json['type']}"
            raise ValueError(msg) from err

        if type_ == TransformType.PIVOT:
            return PivotTransformConfig.create(config_json)
        else:
            return JoinTransformConfig.create(config_json)

    def validate(self, _: Any) -> None:
        pass


@dataclass
class PivotTransformConfig(TransformConfig):
    type_: TransformType
    dataset: str
    path: str
    value_path: str
    label_path: str

    @staticmethod
    def create(config_json: dict[str, Any]) -> PivotTransformConfig:
        for key in ("dataset", "path", "value_path", "label_path"):
            if key not in config_json:
                msg = f"Must specify '{key}' in pivot transform config"
                raise ValueError(msg)

        if config_json["path"] not in config_json["value_path"]:
            msg = f"Invalid pivot transform - path ({config_json['path']}) must be a subpath of value_path ({config_json['value_path']})."
        if config_json["path"] not in config_json["label_path"]:
            msg = f"Invalid pivot transform - path ({config_json['path']}) must be a subpath of label_path ({config_json['label_path']})."

        return PivotTransformConfig(
            type_=TransformType.PIVOT,
            dataset=config_json["dataset"],
            path=config_json["path"],
            value_path=config_json["value_path"],
            label_path=config_json["label_path"],
        )

    def validate(self, config: Any) -> None:
        if not isinstance(config, DatasetConfig):
            msg = f"Must provide DatasetConfig to PivotTransformConfig.validate, got '{type(config)}'"
            raise ValueError(msg)

        if self.value_path not in config.path_to_config:
            msg = f"Missing column config for pivot transform value_key: {self.value_path}"
            raise ValueError(msg)

        if self.label_path not in config.path_to_config:
            msg = f"Missing column config for pivot transform label_key: {self.label_path}"
            raise ValueError(msg)

        value_config = config.path_to_config[self.value_path]
        label_config = config.path_to_config[self.label_path]
        if f"${label_config.name}$" not in value_config.name:
            msg = f"Invalid value column config for pivot transform - must have label column label ('${label_config.name}$') in value column name, got '{value_config.name}'."
            raise ValueError(msg)

        if label_config.include:
            msg = f"Invalid label column config for pivot transform - label column ('{self.label_path}') must not be included in final result."
            raise ValueError(msg)


@dataclass
class JoinTransformConfig(TransformConfig):
    type_: TransformType
    dataset_1: str
    dataset_2: str
    join_key_1: str
    join_key_2: str

    @staticmethod
    def create(config_json: dict[str, Any]) -> JoinTransformConfig:
        for key in ("dataset_1", "dataset_2", "join_key_1", "join_key_2"):
            if key not in config_json:
                msg = f"Must specify '{key}' in join transform config"
                raise ValueError(msg)
        return JoinTransformConfig(
            type_=TransformType.JOIN,
            dataset_1=config_json["dataset_1"],
            dataset_2=config_json["dataset_2"],
            join_key_1=config_json["join_key_1"],
            join_key_2=config_json["join_key_2"],
        )

    def validate(self, datasets: Any) -> None:
        if not isinstance(datasets, dict):
            msg = f"Must provide dictionary of DatasetConfigs to PivotTransformConfig.validate, got '{datasets}'"
            raise ValueError(msg)

        if self.dataset_1 not in datasets:
            msg = f"Invalid dataset_1 in JOIN transform: {self.dataset_1}"
            raise ValueError(msg)
        if self.dataset_2 not in datasets:
            msg = f"Invalid dataset_2 in JOIN transform: {self.dataset_1}"
            raise ValueError(msg)


@dataclass
class ColumnConfig:
    name: str
    path: str
    include: bool
    required: bool

    @staticmethod
    def create(config_json: dict[str, Any]) -> ColumnConfig:
        if "path" not in config_json:
            msg = "Must specify 'path' in dataset config"
            raise ValueError(msg)
        path = str(config_json["path"])
        return ColumnConfig(
            path=path,
            name=config_json.get("name", path.replace("/", ".")),
            include=config_json.get("include", True),
            required=config_json.get("required", False),
        )

    @property
    def has_labels(self) -> bool:
        return bool(re.findall(r"\$([^\$]*)\$", self.name))


@dataclass
class DatasetConfig:
    name: str
    is_metadata: bool
    include: bool
    columns: list[ColumnConfig]
    path_to_transform: dict[str, list[TransformConfig]]
    path_to_config: dict[str, ColumnConfig]
    column_names: list[str]

    def __init__(
        self,
        name: str,
        is_metadata: bool,  # noqa: FBT001
        include: bool,  # noqa: FBT001
        columns: list[ColumnConfig],
        path_to_transform: dict[str, list[TransformConfig]],
    ) -> None:
        self.name = name
        self.is_metadata = is_metadata
        self.include = include
        self.columns = columns
        self.path_to_transform = path_to_transform
        self.path_to_config = {
            column_config.path: column_config for column_config in self.columns
        }
        self.column_names = [column_config.name for column_config in self.columns]
        self._validate()

    @staticmethod
    def create(
        config_json: dict[str, Any],
        transforms: list[TransformConfig],
    ) -> DatasetConfig:
        if "name" not in config_json:
            msg = "Must specify 'name' in dataset config"
            raise ValueError(msg)
        if "columns" not in config_json or not isinstance(config_json["columns"], list):
            msg = "Must specify list 'columns' in dataset config"
            raise ValueError(msg)

        # For columns without name specified, use name of leaf element, if unique
        columns = [ColumnConfig.create(cj) for cj in config_json["columns"]]
        leaf_column_names = [Path(column.path).name for column in columns]
        for column, column_config in zip(columns, config_json["columns"], strict=True):
            if "name" in column_config:
                continue
            leaf_column_name = Path(column.path).name
            if leaf_column_name != column.name:
                if leaf_column_names.count(leaf_column_name) == 1:
                    column.name = leaf_column_name.title()

        path_to_transform: defaultdict[str, list[TransformConfig]] = defaultdict(list)
        for transform in transforms:
            path = getattr(transform, "path", None)
            if path is None:
                msg = "Invalid dataset transform, missing 'path' value."
                raise ValueError(msg)
            path_to_transform[path].append(transform)

        return DatasetConfig(
            name=config_json["name"],
            is_metadata=config_json.get("is_metadata", False),
            include=config_json.get("include", True),
            columns=columns,
            path_to_transform=dict(path_to_transform),
        )

    def _validate(self) -> None:
        # Assert all column names are different
        names = [column_config.name for column_config in self.columns]
        if not len(names) == len(set(names)):
            duplicate_names = {name for name in names if names.count(name) > 1}
            msg = f"Error parsing dataset config, all columns must have unique names, duplicate names: {duplicate_names}"
            raise ValueError(msg)

        # Assert all paths are different
        paths = [column_config.path for column_config in self.columns]
        if not len(paths) == len(set(paths)):
            duplicate_paths = {path for path in paths if paths.count(path) > 1}
            msg = f"Error parsing dataset config, all columns must have unique paths, duplicate paths: {duplicate_paths}"
            raise ValueError(msg)

        for transforms in self.path_to_transform.values():
            for transform in transforms:
                transform.validate(self)

    def add_column(self, column_config: ColumnConfig) -> None:
        self.columns.append(column_config)
        self.path_to_config[column_config.path] = column_config
        self.column_names.append(column_config.name)

    def replace_column_names(
        self, column_name: str, new_column_names: list[str]
    ) -> None:
        replace_index = self.column_names.index(column_name)
        self.column_names = (
            self.column_names[:replace_index]
            + new_column_names
            + self.column_names[replace_index + 1 :]
        )

    def get_column_config(self, path: str) -> ColumnConfig | None:
        # If not columns are specified, include everything with default config
        if not self.path_to_config:
            return ColumnConfig.create({"path": path})
        return self.path_to_config.get(path, None)


@dataclass
class MapperConfig:
    datasets: dict[str, DatasetConfig]
    transforms: list[TransformConfig]

    @staticmethod
    def create(config_json: dict[str, Any] | None = None) -> MapperConfig:
        # If dataset config is not provided, create a "default config" with no columns specified. This will
        # cause all values in the json to be included.
        config_json = config_json or {"datasets": [{"name": "dataset", "columns": []}]}
        if "datasets" not in config_json:
            msg = "Must specify 'datasets' in mapper config"
            raise ValueError(msg)

        # Parse transforms, split into transforms specific to one dataset and those that apply over multiple datasets.
        mapper_transforms: list[TransformConfig] = []
        dataset_transforms: defaultdict[str, list[TransformConfig]] = defaultdict(list)
        for transform in [
            TransformConfig.create(transform_config_json)
            for transform_config_json in config_json.get("transforms", [])
        ]:
            if isinstance(transform, PivotTransformConfig):
                dataset_transforms[transform.dataset].append(transform)
            elif isinstance(transform, JoinTransformConfig):
                mapper_transforms.append(transform)
            else:  # NOTE: should not be possible.
                msg = f"Invalid transform: {transform}"
                raise ValueError(msg)

        # Parse dataset configs.
        datasets: dict[str, DatasetConfig] = {}
        for dataset_config_json in config_json["datasets"]:
            if not isinstance(dataset_config_json, dict):
                msg = f"Invalid dataset config, expected dictionary, got: {dataset_config_json}."
                raise ValueError(msg)
            dataset_config = DatasetConfig.create(
                dataset_config_json,
                dataset_transforms.get(dataset_config_json.get("name", None), []),
            )
            datasets[dataset_config.name] = dataset_config

        for transform in mapper_transforms:
            transform.validate(datasets)

        return MapperConfig(
            datasets=datasets,
            transforms=mapper_transforms,
        )



from dataclasses import dataclass

from allotropy.allotrope.converter import add_custom_information_document
from allotropy.allotrope.models.adm.cell_counting.benchling._2023._11.cell_counting import (
    CellCountingAggregateDocument,
    CellCountingDetectorDeviceControlAggregateDocument,
    CellCountingDetectorMeasurementDocumentItem,
    CellCountingDocumentItem,
    DataProcessingDocument,
    DataSystemDocument,
    DeviceControlDocumentItemModel,
    DeviceSystemDocument,
    ErrorAggregateDocument,
    ErrorDocumentItem,
    MeasurementAggregateDocument,
    Model,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCell,
    TQuantityValueMicroliter,
    TQuantityValueMicrometer,
    TQuantityValueMillionCellsPerMilliliter,
    TQuantityValuePercent,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import JsonFloat
from allotropy.allotrope.schema_mappers.schema_mapper import SchemaMapper
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.parsers.utils.values import quantity_or_none


@dataclass(frozen=True)
class Error:
    error: str
    feature: str | None = None


@dataclass
class Measurement:
    # Metadata
    measurement_identifier: str
    timestamp: str
    sample_identifier: str

    # Measurements
    viability: float
    viable_cell_density: JsonFloat

    # Optional metadata
    batch_identifier: str | None = None
    group_identifier: str | None = None
    sample_draw_time: str | None = None
    written_name: str | None = None
    processed_data_identifier: str | None = None

    # Optional settings
    cell_type_processing_method: str | None = None
    minimum_cell_diameter_setting: float | None = None
    maximum_cell_diameter_setting: float | None = None
    cell_density_dilution_factor: float | None = None
    sample_volume_setting: float | None = None

    # Optional measurements
    viable_cell_count: float | None = None
    total_cell_count: float | None = None
    total_cell_density: JsonFloat | None = None
    dead_cell_count: float | None = None
    dead_cell_density: JsonFloat | None = None

    average_total_cell_diameter: JsonFloat | None = None
    average_live_cell_diameter: float | None = None
    average_dead_cell_diameter: float | None = None
    average_total_cell_circularity: float | None = None
    average_viable_cell_circularity: float | None = None

    average_compactness: float | None = None
    average_area: float | None = None
    average_perimeter: float | None = None
    average_segment_area: float | None = None
    total_object_count: float | None = None
    standard_deviation: float | None = None
    aggregate_rate: float | None = None

    errors: list[Error] | None = None

    # customer information document fields
    debris_index: float | None = None
    cell_aggregation_percentage: JsonFloat | None = None


@dataclass(frozen=True)
class MeasurementGroup:
    measurements: list[Measurement]
    analyst: str | None = None


@dataclass
class Metadata:
    device_type: str
    detection_type: str | None = None
    model_number: str | None = None
    software_name: str | None = None
    file_name: str | None = None
    unc_path: str | None = None
    equipment_serial_number: str | None = None
    software_version: str | None = None
    product_manufacturer: str | None = None
    brand_name: str | None = None
    asset_management_identifier: str | None = None
    device_identifier: str | None = None
    description: str | None = None


@dataclass
class Data:
    metadata: Metadata
    measurement_groups: list[MeasurementGroup]


def has_value(model: object) -> bool:
    return any(value is not None for value in model.__dict__.values())


class Mapper(SchemaMapper[Data, Model]):
    MANIFEST = "http://purl.allotrope.org/manifests/cell-counting/BENCHLING/2023/11/cell-counting.manifest"

    def map_model(self, data: Data) -> Model:
        return Model(
            field_asm_manifest=self.MANIFEST,
            cell_counting_aggregate_document=CellCountingAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    model_number=data.metadata.model_number,
                    product_manufacturer=data.metadata.product_manufacturer,
                    brand_name=data.metadata.brand_name,
                    asset_management_identifier=data.metadata.asset_management_identifier,
                    device_identifier=data.metadata.device_identifier,
                    description=data.metadata.description,
                    equipment_serial_number=data.metadata.equipment_serial_number,
                ),
                data_system_document=DataSystemDocument(
                    file_name=data.metadata.file_name,
                    UNC_path=data.metadata.unc_path,
                    software_name=data.metadata.software_name,
                    software_version=data.metadata.software_version,
                    ASM_converter_name=self.converter_name,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                cell_counting_document=[
                    self._get_technique_document(measurement_group, data.metadata)
                    for measurement_group in data.measurement_groups
                ],
            ),
        )

    def _get_technique_document(
        self, measurement_group: MeasurementGroup, metadata: Metadata
    ) -> CellCountingDocumentItem:
        return CellCountingDocumentItem(
            analyst=measurement_group.analyst,
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_document=[
                    self._get_measurement_document(measurement, metadata)
                    for measurement in measurement_group.measurements
                ]
            ),
        )

    def _get_measurement_document(
        self, measurement: Measurement, metadata: Metadata
    ) -> CellCountingDetectorMeasurementDocumentItem:
        return CellCountingDetectorMeasurementDocumentItem(
            measurement_time=self.get_date_time(measurement.timestamp),
            measurement_identifier=measurement.measurement_identifier,
            sample_document=self._get_sample_document(measurement),
            device_control_aggregate_document=CellCountingDetectorDeviceControlAggregateDocument(
                device_control_document=[
                    DeviceControlDocumentItemModel(
                        device_type=metadata.device_type,
                        detection_type=metadata.detection_type,
                        sample_volume_setting=quantity_or_none(
                            TQuantityValueMicroliter, measurement.sample_volume_setting
                        ),
                    )
                ]
            ),
            processed_data_aggregate_document=self._get_processed_data_aggregate_document(
                measurement
            ),
            error_aggregate_document=self._get_error_aggregate_document(
                measurement.errors
            ),
        )

    def _get_sample_document(self, measurement: Measurement) -> SampleDocument:
        # TODO(ASM gaps): we believe these values should be introduced to ASM.
        custom_document = {
            "group identifier": measurement.group_identifier,
            "sample draw time": self.get_date_time(measurement.sample_draw_time)
            if measurement.sample_draw_time
            else None,
        }
        return add_custom_information_document(
            SampleDocument(
                sample_identifier=measurement.sample_identifier,
                batch_identifier=measurement.batch_identifier,
                written_name=measurement.written_name,
            ),
            custom_document,
        )

    def _get_processed_data_aggregate_document(
        self, measurement: Measurement
    ) -> ProcessedDataAggregateDocument:
        # TODO: average values and stddev should be calculated data
        # TODO(ASM gaps): we believe "total object count" and "aggregate rate" should be introduced to ASM.
        custom_document = {
            "average compactness": quantity_or_none(
                TQuantityValueUnitless, measurement.average_compactness
            ),
            "average area": quantity_or_none(
                TQuantityValueUnitless, measurement.average_area
            ),
            "average perimeter": quantity_or_none(
                TQuantityValueMicrometer, measurement.average_perimeter
            ),
            "average segment area": quantity_or_none(
                TQuantityValueUnitless, measurement.average_segment_area
            ),
            "total object count": quantity_or_none(
                TQuantityValueCell, measurement.total_object_count
            ),
            "standard deviation": quantity_or_none(
                TQuantityValueCell, measurement.standard_deviation
            ),
            "aggregate rate": quantity_or_none(
                TQuantityValuePercent, measurement.aggregate_rate
            ),
            "cell aggregation percentage": quantity_or_none(
                TQuantityValuePercent, measurement.cell_aggregation_percentage
            ),
            "debris index": quantity_or_none(
                TQuantityValueUnitless, measurement.debris_index
            ),
        }
        data_processing_document = DataProcessingDocument(
            cell_type_processing_method=measurement.cell_type_processing_method,
            minimum_cell_diameter_setting=quantity_or_none(
                TQuantityValueMicrometer,
                measurement.minimum_cell_diameter_setting,
            ),
            maximum_cell_diameter_setting=quantity_or_none(
                TQuantityValueMicrometer,
                measurement.maximum_cell_diameter_setting,
            ),
            cell_density_dilution_factor=quantity_or_none(
                TQuantityValueUnitless,
                measurement.cell_density_dilution_factor,
            ),
        )
        processed_data_document = ProcessedDataDocumentItem(
            processed_data_identifier=measurement.processed_data_identifier,
            data_processing_document=data_processing_document
            if has_value(data_processing_document)
            else None,
            viability__cell_counter_=TQuantityValuePercent(value=measurement.viability),
            viable_cell_density__cell_counter_=TQuantityValueMillionCellsPerMilliliter(
                value=measurement.viable_cell_density
            ),
            dead_cell_density__cell_counter_=quantity_or_none(
                TQuantityValueMillionCellsPerMilliliter,
                measurement.dead_cell_density,
            ),
            total_cell_count=quantity_or_none(
                TQuantityValueCell, measurement.total_cell_count
            ),
            total_cell_density__cell_counter_=quantity_or_none(
                TQuantityValueMillionCellsPerMilliliter,
                measurement.total_cell_density,
            ),
            average_total_cell_diameter=quantity_or_none(
                TQuantityValueMicrometer,
                measurement.average_total_cell_diameter,
            ),
            average_live_cell_diameter__cell_counter_=quantity_or_none(
                TQuantityValueMicrometer,
                measurement.average_live_cell_diameter,
            ),
            average_dead_cell_diameter__cell_counter_=quantity_or_none(
                TQuantityValueMicrometer,
                measurement.average_dead_cell_diameter,
            ),
            viable_cell_count=quantity_or_none(
                TQuantityValueCell, measurement.viable_cell_count
            ),
            dead_cell_count=quantity_or_none(
                TQuantityValueCell, measurement.dead_cell_count
            ),
            average_total_cell_circularity=quantity_or_none(
                TQuantityValueUnitless,
                measurement.average_total_cell_circularity,
            ),
            average_viable_cell_circularity=quantity_or_none(
                TQuantityValueUnitless,
                measurement.average_viable_cell_circularity,
            ),
        )
        return ProcessedDataAggregateDocument(
            processed_data_document=[
                add_custom_information_document(
                    processed_data_document, custom_document
                )
            ]
        )

    def _get_error_aggregate_document(
        self, errors: list[Error] | None
    ) -> ErrorAggregateDocument | None:
        if not errors:
            return None

        return ErrorAggregateDocument(
            error_document=[
                ErrorDocumentItem(error=error.error, error_feature=error.feature)
                for error in errors
            ]
        )



from collections.abc import Iterator
from itertools import chain

from allotropy.calcdocs.config import CalculatedDataConfig, MeasurementConfig
from allotropy.calcdocs.extractor import Element
from allotropy.calcdocs.view import ViewData
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
    DataSource,
    Referenceable,
)
from allotropy.parsers.utils.uuids import random_uuid_str


class Constructor:
    def iter_measurement_data_sources(
        self, config: MeasurementConfig, elements: list[Element]
    ) -> Iterator[DataSource]:
        for element in elements:
            if element.get_or_none(config.value) is not None:
                yield DataSource(
                    feature=config.name,
                    reference=Referenceable(
                        element.get_str("uuid")
                    ),  # sould be just element
                    value=None,  # should be value
                )

    def iter_calc_doc_data_sources(
        self, config: CalculatedDataConfig, keys: dict[str, str]
    ) -> Iterator[DataSource]:
        item = config.view_data.get_item(keys)
        for sub_keys in item.iter_keys() if isinstance(item, ViewData) else [{}]:
            if calc_doc := self.get_calc_doc(config, {**keys, **sub_keys}):  # type: ignore[dict-item]
                yield DataSource(
                    feature=config.name,
                    reference=calc_doc,
                    value=None,  # should be calc_doc.value
                )

    def get_calc_doc(
        self,
        config: CalculatedDataConfig,
        keys: dict[str, str],
    ) -> CalculatedDocument | None:
        elements = config.view_data.get_leaf_items(keys)
        value = elements[0].get_float_or_none(config.value)

        if value is None:
            return None

        return CalculatedDocument(
            uuid=random_uuid_str(),
            name=config.name,
            value=value,
            data_sources=[
                data_source
                for source_config in config.source_configs
                for data_source in (
                    self.iter_calc_doc_data_sources(
                        source_config,
                        source_config.view_data.filter_keys(keys),
                    )
                    if isinstance(source_config, CalculatedDataConfig)
                    else self.iter_measurement_data_sources(source_config, elements)
                )
            ],
        )

    def apply_calc_data_config(
        self, config: CalculatedDataConfig
    ) -> list[CalculatedDocument]:
        return [
            calc_doc
            for keys in config.view_data.iter_keys()
            if (calc_doc := self.get_calc_doc(config, keys))
        ]

    def construct(
        self, configs: list[CalculatedDataConfig]
    ) -> list[CalculatedDocument]:
        return list(chain(*[self.apply_calc_data_config(config) for config in configs]))
